// This is part of Skid
//@autor  (P) by Klaus Hammermller, mailto:klaus@ifs.tuwien.ac.at
//@version	1.0 090499
//
//@created	090499  KH
package asgaard.skid;

// Import used asgaard-parts
import asgaard.utils.log.*;
import asgaard.utils.encode.*;

// Import the JDBC classes
import java.sql.*;

// Import used Java classes
import java.util.*;

/**
 * This class avoids the recreation from a set of objects like a database-connection and
 * hides the Database - layer. e.g. if stored procedures are available or not.
 * All the <i>SQL-Code</i> is encapsulated in this class, different SQL-dialects
 * are maintained with the Resource-classes <code>SQLRes*.java</code>.
 *
 *@see asgaard.skid.resources.SQLRes
 */
public class BeanBrokerCall {

  /**
   * Method-name convention prefix: getXXXX - Methods are analyzed
   */
  public static final String NAME_PREFIX = "get";
  /**
   * The prefix of the table carrying the Keys of an Bean
   */
  private static final String KEYTABLE_PREFIX = "K";
  /**
   * The prefix of the table carrying the Keys of an Class
   */
  private static final String CLASSTABLE_PREFIX = "C";
  /**
   * The prefix of an Keytable-Field (field are named by the Method-Index).
   */
  private static final String FIELD_PREFIX = "F";
  /**
   * The prefix of the primary Bean-ID Field Keytable-Field.
   */
  private static final String PRIMARY_KEY_FIELDNAME = "p";
  /**
   * The prefix of the secondary Fieldnumber Keytable-Field.
   */
  private static final String SECONDARY_KEY_FIELDNAME = "f";

  private Statement stmt;             // The encapsuled Statement
  private ResultSet rset = null;      // The ResultSet (if any)
  private String[] stmtTok;           // The simple statement
  private String[] paramTok;          // Names of the Parameter
  private int[] tokIndx;
  private boolean prep;               // whether the statement is a stored proc.
  private boolean getResult;          // whether the sored proc. produces a ResultSet

  private ObjectEncoder enc = new ObjectEncoder( );                   // set UUEncoder
  private ObjectDecoder dec = new ObjectDecoder( );                   // set UUDecoder

  private static String srvClass;
  private final static boolean debug = Skid.debug;
  protected static boolean blobWorkaround = false;

  protected static ResourceBundle sRes;      // The SQL-syntax resources
  private static BeanBroker broker;              // The automated Loging
  private static Connection conn;            // The Connection

  /**
   * Creeates a new database-call.
   *@param Conn the database - connection
   *@param Log  the Log for automated backup
   *@param sqlRes the SQL-Syntax-Resources
   */
  public static void setConnection(Connection aConn, BeanBroker aBroker, ResourceBundle sqlRes) {
    sRes = sqlRes;
    conn = aConn;
    broker = aBroker;
  }

  /**
   * Creeates a new database-call.
   *@param callStr  the Statement (parameterless)
   *@param stmtStr  the Log-Statement for automated backup (without stored procedure call)
   *@param getResult if the preprocessed statement produces a ResultSet or not
   *@param createStr is the <code>CREATE TABLE</code> Statement to analyze the parameter names (optional).
   *@throws SQLException if the statement cannot be created.
   */
  public BeanBrokerCall(String callStr, String stmtStr, String createStr, boolean getResult) throws SQLException {
    this.prep = !((callStr == null) || (callStr.equals("")));
    StringTokenizer tokens = new StringTokenizer( stmtStr, "}", false );
    stmtTok = new String[tokens.countTokens()];
    tokIndx = new int[stmtTok.length-1];
		for (int i=0;i<stmtTok.length;i++) {
		  stmtTok[i] = tokens.nextToken();
      int l = stmtTok[i].indexOf("{");
      if (l > 0) {
        tokIndx[i] = Integer.parseInt( stmtTok[i].substring(l+1, stmtTok[i].length()) );
        stmtTok[i] = stmtTok[i].substring(0, l);
      }
    }
    this.getResult = getResult;
    srvClass = this.getClass().getName();
    if (createStr != null) {                      // generating the attribute-name-list
      StringTokenizer pTokens = new StringTokenizer( createStr.substring( createStr.indexOf( "(" ) + 1, createStr.lastIndexOf( ")" )), "," );
      paramTok = new String[pTokens.countTokens()];
      for (int j=0;j<paramTok.length;j++) {
        String s = pTokens.nextToken().trim();
        paramTok[j] = s.substring(0, s.indexOf( " " ));
      }
    }
    if (prep)
      stmt = conn.prepareCall (callStr);          // create a stored procedure call
    else
      stmt = conn.createStatement();              // Create a statement
  }

  /**
   * Sends the LogStatements
   */
  private boolean logMsg(String srvClass, String callPoint, String aStmt, boolean debugmsg) {
    try {
      if (debugmsg)
        return broker.logMsg(srvClass, callPoint, LogMsg.LOG_DEBUG, true, "exec.SQL", aStmt);
      else
        return broker.logMsg(srvClass, callPoint, LogMsg.LOG_BACKUP, true, "exec.SQL", aStmt);
    } catch (Exception e) {
      return false;
    }
  }

  /**
   * Does the wraped call including a backup (if LogLevel >= BACKUP).
   * The number of required parameters is not checked!
   *
   *@param args a set of Arguments as <code>String</code>, for Backup all none-Strings are encoded to HEX.
   *@param callPoint the Servicepoint for the call (used for the LogMsg)
   *@throws SQLException if the statement cannot be processed
   */
  public ResultSet exec (Object[] args, String callPoint) throws SQLException {
    rset = null;
    String aStmt = stmtTok[0];
    if (! prep) {                                           // shall we do the work producing a backup message?
      if (args.length > 0) {                                // Args
        aStmt = new String( stmtTok[0] );
        for (int i=0;i<tokIndx.length;i++)
          aStmt += bean2string(args[tokIndx[i]]) + stmtTok[i+1];
      }
      logMsg(srvClass, callPoint, aStmt, getResult);        // put the Backup Statement
    }
    if (prep) {                                             // do a prepared Statement ...
      CallableStatement sp = (CallableStatement) stmt;
      int k = 1;
      if (blobWorkaround)                                   // WORKAROUNT vvvvvvvvvvvvvvvvvv
        for ( int i=0;i<args.length; ++i )	                // Parameter define
          if (args[i] != null) {
            sp.setObject(k, bean2string(args[i]));          // <-- encode binary
            k++;
          }
      else                                                  // NORMAL vvvvvvvvvvvvvvvvvv
        for ( int j=0;i<args.length; ++j )	                // Parameter define
          if (args[j] != null) {
            sp.setObject(k, args[j]);
            k++;
          }
	    if ( getResult ) {					                          // ResultSet returned
        rset = sp.executeQuery();
        return rset;
	    } else {
        sp.execute ();					                            // StalledProc call
        return null;
    	}
    } else
        rset = stmt.executeQuery (aStmt);                   // ... or a normal query
        return rset;
  }

  /**
   * Returns the Object containing to the column-number and the actual resultset,
   * hides any SQL-Exception.
   *
   *@param column number of the field, first starts with 1
   *@param isBean if the Object may be encoded
   *@return the Object if any
   */
  public Object getObject(int column, boolean isBean) {
    if (rset == null) return null;
    try {
      if (isBean && (blobWorkaround || (! prep)))
        return string2bean((String) rset.getObject( column ));
      else
        return rset.getObject( column );
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return null;
    }
  }

  /**
   * Returns an <code>int</code> containing to the column-number and the actual resultset,
   * hides any SQL-Exception.
   *
   *@param column number of the field, first starts with 1
   *@return an int-value (-1 if it fails)
   */
  public int getInt(int column) {
    if (rset == null) return -1;
    try {
      return rset.getInt( column );
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return -1;
    }
  }

  /**
   * Returns an <code>String</code> containing to the column-number and the actual resultset,
   * hides any SQL-Exception.
   *
   *@param column number of the field, first starts with 1
   *@return an String (<code>null</code> if it fails)
   */
  public String getString(int column) {
    if (rset == null) return null;
    try {
      return rset.getString( column );
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return null;
    }
  }

  /**
   * Returns an <code>long</code> containing to the column-number and the actual resultset,
   * hides any SQL-Exception.
   *
   *@param column number of the field, first starts with 1
   *@return an long-value (-1 if it fails)
   */
  public long getLong(int column) {
    if (rset == null) return -1;
    try {
      return rset.getLong( column );
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return -1;
    }
  }

  /**
   * Indicates, if there is a next Entry at the Resultset,
   * hides any SQL-Exception.
   *@return if ther is a former result
   */
  public boolean next() {
    if (rset == null) return false;
    try {
      return rset.next();
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return false;
    }
  }


  /**
   * Returns a new Element in the Filter-clause.
   *@param fieldno number of the field (starts with 0)
   *@param filter the filter-code
   *@param value the value to compare (to a <code>String</code> quotes are added)
   *@param where the old clause
   *@return a new clause
   */
  public String getFilterClause(byte fieldno, byte filter, Object value, String where) {
    String s;
    if (value instanceof String) {
      s = "'" + value + "'";
    } else {
      s = value.toString();
    }
    s = getSQLOp(filter, "(" + paramTok[fieldno], s + ")");
    if (s.length() > 0) {
      if (where.length() > 0)
        return where + " AND " + s;
      else
        return s;
    } else
      return where;
  }

  /**
   * Generates the Tablename for the keys of a Class
   *@returns an valid Tablename for a Class
   */
  public static String getTableName(int classKey) {
    return KEYTABLE_PREFIX + String.valueOf(classKey);
  }

  /**
   * Returns an "null-Clause" for the Search in a Keytable
   *@returns the Clause <code>p = 0</code>
   */
  public static String getKeyNullClause() {
    return PRIMARY_KEY_FIELDNAME + " = 0";
  }
  /**
   * Returns the Create-Clause for the Key(s) in the keytable
   *@param Class the class generate the key-clause
   *@returns the clause
   */
  public static String getCreateKeyClause(Class aClass) {
    try {
      if (aClass.newInstance() instanceof RemoteBean)
        return PRIMARY_KEY_FIELDNAME + " " + sRes.getString("INTEGER");
      else
        return PRIMARY_KEY_FIELDNAME + " " + sRes.getString("INTEGER") + ", " +
               SECONDARY_KEY_FIELDNAME + " " + sRes.getString("INTEGER");
    } catch (Exception e) {
      return PRIMARY_KEY_FIELDNAME + " " + sRes.getString("INTEGER") + ", " +
             SECONDARY_KEY_FIELDNAME + " " + sRes.getString("INTEGER");
    }
  }

  /**
   * Returns the Clause with an prefix - Semikolon
   *@param clause the token
   *@returns <code>', clause'</code>
   */
  public static String addListClause(String clause) {
    return ", " + clause;
  }

  /**
   * Returns the Clause with an prefix - AND
   *@param clause the token
   *@returns <code>' AND clause'</code>
   */
  public static String addAndClause(String clause) {
    return " AND " + clause;
  }

  /**
   * Returns the Key-Clause for the second Key
   *@param fieldno the number of the property
   *@returns the clause
   */
  public static String getSecondaryKeyClause(int fieldno) {
    return " AND " + SECONDARY_KEY_FIELDNAME + " = " + String.valueOf( fieldno );
  }

  /**
   * Returns a part of a CREATE statement to construct the Key-Table of an Bean.
   *@param fieldnr number of the getXXX-method (by <code>Class.getMethods</code>)
   *@param type of the Return-Type of the method
   *@param part the old token
   *@return the concated Tokens
   */
  public static String getCreateKeyToken(int fieldnr, byte type, String part) {
    if (part.length() > 0)
      part += ", ";
    switch (type) {
      case BeanKeyCall.BOOLEAN: case BeanKeyCall.BYTE:
        return part + FIELD_PREFIX + String.valueOf(fieldnr) + " " + sRes.getString("TINYINT");
      case BeanKeyCall.DOUBLE:
        return part + FIELD_PREFIX + String.valueOf(fieldnr) + " " + sRes.getString("DOUBLE");
      case BeanKeyCall.FLOAT:
        return part + FIELD_PREFIX + String.valueOf(fieldnr) + " " + sRes.getString("FLOAT");
      case BeanKeyCall.SHORT:
        return part + FIELD_PREFIX + String.valueOf(fieldnr) + " " + sRes.getString("SMALLINT");
      case BeanKeyCall.INT: // case BeanKeyCall.BEAN:
        return part + FIELD_PREFIX + String.valueOf(fieldnr) + " " + sRes.getString("INTEGER");
      default:
        return part + FIELD_PREFIX + String.valueOf(fieldnr) + " " + sRes.getString("BIGINT");
    }
  }

  /**
   * Creates a Token for selecting from a Key-Table.
   */
  public static String getSelectKeyToken(int fieldnr, byte op, String value) {
    return getSQLOp(op, FIELD_PREFIX + String.valueOf(fieldnr), value);
  }

  /**
   * Returns a Filter-clause > 0.
   *@param fieldno number of the field (starts with 0)
   */
  public String getZeroClause(byte fieldno) {
    return "(" + paramTok[fieldno] + " > 0)";
  }

  /**
   * Returns the Operater in SQL-syntax defined by <code>BeanBroker.FILTER_*</code> adding
   * an prefix- and postfix String.
   *@param filter the filter-code
   *@param prefix the prefix
   *@param postfix the postfix
   *@return the String with the operator
   */
  protected static String getSQLOp(byte filter, String prefix, String postfix) {
    switch (filter) {
      case BeanBroker.FILTER_EQ: return prefix+" = "+postfix;
      case BeanBroker.FILTER_GR: return prefix+" > "+postfix;
      case BeanBroker.FILTER_SM: return prefix+" < "+postfix;
      case BeanBroker.FILTER_GR_EQ: return prefix+" >= "+postfix;
      case BeanBroker.FILTER_SM_EQ: return prefix+" <= "+postfix;
      case BeanBroker.FILTER_NOT_EQ: return prefix+" <> "+postfix;
      default: return "";
    }
  }

  /**
   * Converts a Bean to a String
   */
  private String bean2string(Object bean) {
    if (bean == null) return "";
    if (bean instanceof java.lang.String)
      return (String) bean;
    try {
      enc.setObject( bean );		                                       //   ... und ab!
      return enc.getEncoded();                                         // ... it's magic
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return "";
    }
  }

  /**
   * Converts a String to a Bean
   */
  private Object string2bean(String bean) {
    if (bean == null) return "";
    try {
      dec.setEncoded( bean );		                                       //   ... und ab!
      return dec.getObject();                                          // ... it's magic
    } catch (Exception e) {
      System.err.println( e.toString());
      if (debug) e.printStackTrace();
      return "";
    }
  }
}