// This class is part of Skid
//@autor  (P) by Klaus Hammermller, mailto:klaus@ifs.tuwien.ac.at
//@version	1.0 140499
//
//@created	140499  KH reimplemented (was former part of DataObject)
package asgaard.skid;
// Import the java classes used
import java.io.*;

/**
 * This class models one Relationship between one ore two Beans. It is able to
 * maintain it and carrys all nessecary keys. Additionally it can be used to iterate
 * throug the Bean-Ontology.
 */
public class BeanLink implements Serializable {
  /**
   * Link to the BeanBroker (one per VM).
   */
  private transient static RemoteBroker broker = BeanBroker.THIS;
  /**
   * Update Policy for saving modifications.
    *@see asgaard.skid.SkidBean.checkModification
   */
  public static byte update_policy = BeanBroker.UPDATE_WEAK;

  protected int From = 0;                                  // 0 FROM
  protected int Order = 0;                                 // 1 (not part of the key!)
  protected int To = 0;                                    // 2 TO
  protected int By = 0;                                    // 3 RELATE
  protected long Created = System.currentTimeMillis()-1;   // 4 timestamp
  protected long Modified = Created+1;                     // 5 timestamp
  protected long Age = Created;
  protected long Begin = 0;                                // 6 timestamp
  protected long End = 0;                                  // 7 timestamp
  protected int State = 0;                                 // 8 read/add/mod/exec for system/owner/creator/group/other

  private static final boolean debug = false;
    /**
     * Creates an empty new Link-object.
     */
    public BeanLink() {  }

    /**
     * Creates a new Link-object.
     *@param from key of the origin Bean
     *@param to key of the target Bean
     *@param  by key of the link-Bean (kind of Link)
     *@param  begin timestamp of the start of the relationship
     *@param  end timestamp of the end of the relationship
     *@param  state access-rights to the target Bean about this relationship
     */
    public BeanLink(int from, int to, int by, long begin, long end, int state) {
      From = from;
      To = to;
      By = by;
      Begin = begin;
      End = end;
      State = state;
    }

    /**
     * Creates a new Link-object.
     *@param from key of the origin Bean
     *@param to key of the target Bean
     *@param  by key of the link-Bean (kind of Link)
     *@param  begin timestamp of the start of the relationship
     *@param  end timestamp of the end of the relationship
     *@param  state access-rights to the target Bean about this relationship
     *@throws Exception if an Bean is not valid.
     */
    public BeanLink(RemoteBean from, RemoteBean to, RemoteBean by, long begin, long end, int state) throws Exception {
      if (from == null)
        From = 0;
      else
        From = from.getKey();
      if (to == null)
        To = 0;
      else
        To = to.getKey();
      if (by == null)
        By = 0;
      else
        By = by.getKey();
      Begin = begin;
      End = end;
      State = state;
    }
    /**
     * Creates a new Link-object.
     *@param from key of the origin Bean
     *@param  order a local order by an ontology to from
     *@param to key of the target Bean
     *@param  by key of the link-Bean (kind of Link)
     *@param  created timestamp when the link was created
     *@param  mod timestamp when the key was last modified
     *@param  begin timestamp of the start of the relationship
     *@param  end timestamp of the end of the relationship
     *@param  state access-rights to the target Bean about this relationship
     */
    public BeanLink(int from, int order, int to, int by, long create, long mod, long begin, long end, int state) {
      From = from;
      Order = order;
      To = to;
      By = by;
      Created = create;
      Modified = mod;
      Age = Modified;
      Begin = begin;
      End = end;
      State = state;
    }

    /**
     * Has the Relation been modified?
     */
    public boolean isChanged() {
      return (Age != Modified);
    }

    /**
     * Mark the Relation as modified.
     */
    public void modify() {
      Modified = System.currentTimeMillis();
    }

    /**
     * Set's the Age to the current (modified) timestamp
     */
    public void setAge() {
      Age = Modified;
    }
    /**
     * Does all the saving.
     *@throws Exception if the save failed.
     */
    public void flush() throws Exception {
      if (broker == null)
        throw new SkidException("Link not bound to any broker.");
      if (isChanged())
        broker.putLink(this);
      setAge();
    }

    /**
     * Returns the From-Bean in this Relation.
     *@return the bean if there is any
     */
    public RemoteBean getFrom() {
      try {
        return broker.getBean( From );
      } catch (Exception e) {
        if (debug) {
          System.err.println(e.toString());
          e.printStackTrace();
        }
        return null;
      }
    }

    /**
     * Returns the To-Bean in this Relation.
     *@return the bean if there is any
     */
    public RemoteBean getTo() {
      try {
        return broker.getBean( To );
      } catch (Exception e) {
        if (debug) {
          System.err.println(e.toString());
          e.printStackTrace();
        }
        return null;
      }
    }

    /**
     * Returns the Relation-Bean.
     *@return the bean if there is any
     */
    public RemoteBean getBy() {
      try {
        return broker.getBean( By );
      } catch (Exception e) {
        if (debug) {
          System.err.println(e.toString());
          e.printStackTrace();
        }
        return null;
      }
    }

    /**
     * Returns the primary Key of the Relation.
     *@return a String with the tripple "from.to.by"
     */
    public String getKey() {
      return getKey( From, To, By );
    }
    /**
     * Returns a primary Key of the Relation with the given keys.
     *@param from key of the origin Bean
     *@param to key of the target Bean
     *@param  by key of the link-Bean (kind of Link)
     *@return a String with the tripple "from.to.by"
     */
    public static String getKey(int from, int to, int by) {
      return String.valueOf( from ) + "." + String.valueOf( to ) + "." + String.valueOf( by );
    }

    /**
     * Returns a Filter-Object corresponding to this Key.
     *@return the filterobject
     */
    public BeanLinkFilter getFilter() {
      return new BeanLinkFilter( this );
    }

}
