// This class is part of Skid
//@autor  (P) by Klaus Hammermller, mailto:klaus@ifs.tuwien.ac.at
//@version	1.0 270499
//
//@created	270499  KH
package asgaard.skid;

// Import Skid-Api
import asgaard.skid.*;
import asgaard.utils.log.*;

// Import the java classes used
import java.io.*;

//import java.rmi.*;

import java.lang.reflect.*;


/**
 * This Class is a special lightweight type of Bean which does not carry any
 * Attributes but it's name (no multilanguage-support). Therefor it is used to hold e.g. String -
 * identifyers used by the System.
 */
public class LightBean implements Serializable, RemoteBean {

  /**
   * Link to the RemoteBroker (one per VM).
   */
  protected transient static RemoteBroker broker = BeanBroker.THIS;
  /**
   * Update Policy for saving modifications.
   *@see asgaard.skid.SkidBean.save
   *@see asgaard.skid.SkidBean.checkModification
   */
  public static byte update_policy = BeanBroker.UPDATE_ALL;

  /**
   * The History - Policy for saving UNDO-information.
   * Valid values are <code>HISTORY_ALL, HISTORY_NONE</code> and any number >
   * <code>0</code> defining the depth of the history.
   */
  public static int history_policy = BeanBroker.HISTORY_NONE;

//===============================================
// KEYS -----------------------------------------
  protected int key = BeanBroker.ID_FAIL;
  protected int cKey = BeanBroker.ID_FAIL;
  protected long tsCreated = System.currentTimeMillis()-1;
  protected long tsModified = tsCreated+1;
// KEYS -----------------------------------------
//===============================================
  private boolean beanIsReadOnly = false;
  private boolean modified = true;
  private boolean newBean = true;
  protected transient final static boolean debug = Skid.debug;

  /**
   * New Instance
   */
  public LightBean() { }

  /**
   * Initialises a new Lightweight Bean with a given name and timestamps <b>Do not use manually</b>!!!
   *@param key of the Bean
   *@param calssKey of the Bean
   *@param name the short name of the def
   *@param content the long name of the def
   *@param tsC created - timestamp
   *@param tsM modified - timestamp
   *@throws SkidException
   */
  public void init (int key, int classKey, String name, Object content, long tsC, long tsM) throws SkidException {
    this.key = key;
    this.cKey = classKey;
    tsCreated = tsC;
    tsModified = tsM;
  }

  /**
   * Get Link to the RemoteBroker (one per VM).
   */
  public RemoteBroker getBroker() {
    return broker;
  }

  /**
   * Get's the update policy defined by the <code>UPDATE_*</code> constants.
   */
  public byte getUpdatePolicy() {
    return update_policy;
  }

  /**
   * Get the Identifier of the Bean which shall be unique worldwide for different
   * Objects (like different Plans with luckily the same Name) and common for
   * identical Objects (like a Person reentered), <b>do not use manually</b>.
   */
  public String getIdentifier() {
    String classname = this.getClass().getName();
    return classname.substring(classname.lastIndexOf(".")+1, classname.length()) + ":" + getName();
  }

  /**
   * Returns the name of the Bean which shold be human readable and may be
   * unique in some cases, but this is not strictly required.
   */
  public String getName() {
    return "L"+String.valueOf(tsCreated);
  }

  /**
   * (Unique) Primary Key of the Bean, created from the attribute
   * <code>name</code>, the value is negative.
   *@throws SkidException not used.
   */
  public int getKey() throws SkidException {
    if (key == BeanBroker.ID_FAIL)
      throw new SkidException("Key not valid now.");
    return key;
  }

  /**
    * Returns the type of the Bean's class, used for the loading-signature.
   *@throws SkidException
   */
  public int getClassKey() throws SkidException {
    if (cKey == BeanBroker.ID_FAIL)
      throw new SkidException("Key not valid now.");
    return cKey;
  }

  /**
   * Returns the type of the Bean, used for the loading-signature.
   *@throws RemoteException
   */
  public byte getBeanType() {
//  public byte getBeanType() throws RemoteException {
    return BeanBroker.TYPE_LIGHTWEIGHT;
  }

  /**
   * Timestamp "Created" of the Bean. Based on System.currentTimeMillis().
   *@see System#currentTimeMillis
   *@throws SkidException not used
   */
  public long getCreated() throws SkidException {
    return tsCreated;
  }

  /**
   * Timestamp "Modified" of the Bean. Based on System.currentTimeMillis().
   *
   *@see System#currentTimeMillis
   *@throws SkidException not used
   */
  public long getModified() throws SkidException {
    return tsModified;
  }

  /**
   * Timestamp "Age" of the Bean (= current version by the Modified-Timestamp).
   * Based on System.currentTimeMillis().
   *
   *@see System#currentTimeMillis
   *@throws SkidException not used
   */
  public long getAge() throws SkidException {
    if (modified)
      return tsModified - 1;
    else
      return tsModified;
  }

  /**
   * Sets the Age to the Modified timestamp, therefor IsChanged = false,
   * Intended usage only from RemoteBroker.
   *
   *@see asgaard.skid.RemoteBroker#put
   */
  public void setAge() {
    modified = false;
    newBean = false;
  }

  /**
   * Returns a set of Links defined by an filterobject.
   * If this filterobject is missing (<code>null</code>) all Links of this Bean are searched.
   *@param filter the filterobject
   *@return a set of links
   *@see asgaard.skid.BeanLinkFilter
   *@throws Exception if the operation failes
   */
  public BeanLink[] getLink(BeanLinkFilter filter) throws Exception {
    BeanLinkFilter f = filter;
    if (f == null) {                                   // prepare new filter if missing...
      f = new BeanLinkFilter();
      f.setFrom( this.getKey(), BeanBroker.FILTER_EQ );
    }
    return broker.getLink( f );
  }
  /**
   * Adds a Relaition between Beans
   *@param link one relation
   *@return wether an update is possible or not.
   *@throws Exception if the key is not valid or anything else fails
   */
  public boolean putLink(BeanLink link) throws Exception {
    return broker.putLink( link );
  }
    /**
     * Adding an Event and replicating it to all suscribed Objects.
     *@param event the kond of the event (not <code>null</code>)
     */
    public void addEvent(RemoteBean event) {
      try {
        broker.addEvent((RemoteBean) this, event);
      } catch (Exception e) {      }
    }

    /**
     * Adding an Event-Pattern and suscribing this type of event.
     *@param producer of the event (not <code>null</code>)
     *@param event the kond of the event (not <code>null</code>)
     */
    public void suscribeEvent(RemoteBean producer, RemoteBean event) {
      try {
        broker.suscribeEvent(producer, event, (RemoteBean) this);
      } catch (Exception e) {      }
    }

    /**
     * Returns the Events for this bean which are newer than the given time
     *@param newer than the given time
     *@returns a set of events
     */
    public SkidEvent[] getEvent(long newer) {
      try {
        return broker.getEvent((RemoteBean) this, newer);
      } catch (Exception e) {
        return new SkidEvent[0];
      }
    }

  /**
   * All Keys of the Bean which are established at the first insert to the
   * RemoteBroker, <b>do not use manually</b>. Only use in <code>insertBean</code>
   *
   *@see asgaard.skid.RemoteBroker.insertBean
   *@param  pKey the generated unique primary key
   *@param  broker  defines the owning broker
   *@return false if the key was set before.
   */
  public boolean init(int pKey, int cKey, RemoteBroker broker) {
      this.key = pKey;
      this.cKey = cKey;
      this.broker = broker;
      modify(null, null);
      setAge();
      return true;
  }
  /**
   * Set the modified flag and savin the old Value for the undo-mechanism (if
   * the history_policy allowes this). To avoid storing information you can use
   * the <code>history_policy</code> or use <code>null, null</code> as parameter
   * values.
   *
   *@param  oldVlaue  the old value (if any)
   *@param  fieldName the name of the field which was changed (<code>see
   * getClass.getDeclaredField()</code>)
   *@see UndoHistory
   */
  protected void modify(Object oldValue, String fieldName) {
    tsModified = System.currentTimeMillis();
    modified = true;
  }

  /**
   * Has the Bean been changed?
   */
  public boolean isChanged() {
    return modified;

  }

  /**
   * Is the Bean new?
   */
  public boolean isNew() {
    return newBean;
  }

  /**
   * Ist the Bean read only?
   */
  public boolean isReadOnly () {

    return beanIsReadOnly || (tsCreated == 0);

  }


  /**
   * Sets (replaces if any before) the given objects.
   *@param contain the wrapped object
   *@throws SkidException if the given Object is not Serializable
   */
  public void setObject(Object contain) throws SkidException {
    if (! (contain instanceof Serializable))
      throw new SkidException("Object not Serializable.");
    modify(contain, "obj");
  }

  /**
   * Returns the wrapped Object, in this case it's the Longname.
   */
  public Object getObject() {
    return null;
  }

  /**
   * Returns a Filter for this (and all parent) class.

   */

  public SkidBeanFilter getFilter() {

    try {

      return new SkidBeanFilter( this.getClass());

    } catch (Exception e) {

      return null;

    }

  }

  /**
   * Returns an array containing Method objects reflecting all the public member

   * methods of the class or interface represented by this Class object,

   * including those declared by the class or interface and and those inherited

   * from superclasses and superinterfaces. Returns an array of length 0 if the

   * class or interface has no public member methods.

   *
   *@throws SecurityException if access to the information is denied.
   */

  public Method[] getMethods() throws SecurityException {

    return this.getClass().getMethods();

  }

  /**
   * Invokes the underlying method represented by this Method object, on the

   * specified object with the specified parameters. Individual parameters are

   * automatically unwrapped to match primitive formal parameters, and both
   * primitive and reference parameters are subject to widening conversions as
   * necessary. The value returned by the underlying method is automatically
   * wrapped in an object if it has a primitive type.
   *
   *@param name the name of the Method
   *@param args the Arguments
   *@param modify call the modify-method too? (not including undo)
   *@return a result returned (if any)
   *@see java.lang.reflect.Method#invoke
   *@throws IllegalAccessException if the underlying method is inaccessible.

   *@throws IllegalArgumentException if the number of actual and formal
   * parameters differ, or if an unwrapping conversion fails.
   *@throws InvocationTargetException if the underlying method throws an
   * exception.
   *@throws NullPointerException if the specified object is null.
   *@throws NoSuchMethodException if there is no Method with this name
   */
  public Object invoke(String name,Object args[], boolean modify) throws IllegalAccessException,

                                                    IllegalArgumentException,

                                                    InvocationTargetException,

                                                    NullPointerException,

                                                    NoSuchMethodException {

    Class[] params = new Class[args.length];

    for (int i=0;i<args.length;i++) {

      params[i] = args[i].getClass();

    }

    Object ret = this.getClass().getMethod(name, params).invoke(this, args);

    if (modify)

      modify(null, null);

    return ret;

  }

  /**
   * Save all modifications (if any) and checks if the update is possible
   * (timestamp of the RemoteBroker). Remove the modified flag, set the
   * modified-timestamp.
   *
   *@throws Exception if the save failed.
   */
  public void flush() throws Exception {
    if (broker == null)
      broker = BeanBroker.THIS;
    if (broker == null)
      throw new SkidException("Bean not bound to any broker.");
    if (debug) broker.logMsg( this.getClass().getName(), "flush#0", LogMsg.LOG_DEBUG, true, "CHANGED: "+String.valueOf(isChanged()));
    if (isChanged()) {
      Object[] keys = broker.putBean(this);
      if (keys==null) {
        if (debug) broker.logMsg( this.getClass().getName(), "flush#1", LogMsg.LOG_DEBUG, keys!=null, "CHANGED: "+String.valueOf(isChanged()));
        throw new SkidException("Newer Bean exists");
      }
      key = ((Integer)keys[0]).intValue();
      cKey = ((Integer)keys[1]).intValue();
      tsModified = ((Long)keys[2]).longValue();
      setAge();
      if (debug) broker.logMsg( this.getClass().getName(), "flush#2", LogMsg.LOG_DEBUG, true, "AGE: "+String.valueOf(getAge()));
    }
  }

  /**
   * Produces a discription of the last changes which can be undone <b>Dummy</b>.
   *@return null
   */
  public String[] getChangeHistory() {
    return null;
  }

  /**
   * Clones nothing <b>Dummy</b>

   *@return null

   */

  public Object cloneBean() throws CloneNotSupportedException {

//  public Object cloneBean() throws CloneNotSupportedException, RemoteException {

    return null;

  }


  /**
   * Undo <b>Dummy - HAS TO BE THE LAST METHOD OF A REMOTEBEAN ROOT-IMPLEMENTATION</b>

   */

  public boolean undo(int steps) {

    return false;

  }


}










