// This class is part of Skid
//@autor  (P) by Klaus Hammermller, mailto:klaus@ifs.tuwien.ac.at
//@version	2.0 080499
//
//@created	080499  KH reimplementation from asgaard.skid.DataObject
package asgaard.skid;

// Import skid-API
import asgaard.lang.*;
import asgaard.utils.log.*;

// Import the java classes used
import java.util.*;
//import java.rmi.*;

//import java.rmi.server.UnicastRemoteObject;

import java.io.*;

import java.lang.reflect.*;

/**
 * Implementation of an Data-Bean.
 *<p>
 * <b><i>Important:</i></b> Once in use, a Class HAS NOT TO BE CHANGED ANY MORE!
 * If a class is changes after use, there will be a serialisation-error getting the bean
 * out of the database.
 *<p>
 * <b><i>Important:</i></b> Key values which are accessible via the
 * <code>RemoteBroker</code> have to be defined following the
 * <i>Java-Beans-Convention</i> using the <code>getXXXX</code>
 * methods for key-generation.
 *<p>
 * <b><i>Important:</i></b> <code>public static RemoteBroker THIS</code> has to be defined
 * within each VM to guarantie regular behavior within Skid!
 *<p>
 * <b><i>Remark:</i></b> All attribues used by the Object which
 * are not intended to be persitent have to be declared <code>transient</code>
 * and restored by default or the parameterless constructor. Within implementing
 * new <code>setXXXX</code> methods call the <code>modify()</code> method to set
 * the new timestamps.
 *<p>
 *<b><i>Minimal Usage:</i></b> (The RemoteBroker should be instanciated at this time)
 *<code>  SkidBean b = new SkidBean();</code><br>
 *<code>  b.setName()                 </code> (optional) <br>
 *<code>  b.flush();</code><br>
 *<p>
 * <i>Remark:</i> A System-Object cannot be modified and carries the created-Timestamp 0
 *
 *@see asgaard.skid.RemoteBroker
 */
public class SkidBean implements Serializable, RemoteBean {
//public class SkidBean extends UnicastRemoteObject implements Serializable, RemoteBean {

  /**
   * Last method of the Interface - Implementation
   */
  public static final String LAST_METHOD = "undo";
  /**
   * Empty key of Bean "def:<none>"
   */
   public static final int EMPTY_KEY = 100;

  /**
   * Link to the RemoteBroker (one per VM).
   */
  protected transient static RemoteBroker broker = BeanBroker.THIS;

  /**
   * The History - Policy for saving UNDO-information.
   * Valid values are <code>HISTORY_ALL, HISTORY_NONE</code> and any number >
   * <code>0</code> defining the depth of the history.
   *
   *@see asgaard.skid.RemoteBean.HISTORY_ALL
   *@see asgaard.skid.RemoteBean.HISTORY_NONE
   *@see modify
   *@see undo
   */
  public static int history_policy = BeanBroker.HISTORY_ALL;

//===============================================
// KEYS -----------------------------------------
  private int key = BeanBroker.ID_FAIL;
  private int cKey = BeanBroker.ID_FAIL;
  private long tsCreated = System.currentTimeMillis()-1;
  private long tsModified = tsCreated+1;
  private long tsAge = tsCreated;
// KEYS -----------------------------------------
//===============================================
  private boolean beanIsReadOnly = false;
  protected transient final static boolean debug = Skid.debug;


  // Storage of the UNDO-Mechanism **************************
  private transient Vector undoVec = new Vector();

  /**
   * Instanciates a Bean and does all the initialisation work for transient
   * attributes.
   */
  public SkidBean() {
  }

  /**
   * Get Link to the RemoteBroker (one per VM).
   */
  public RemoteBroker getBroker() {
    return broker;
  }

  /**
   * Get's the update policy defined by the <code>UPDATE_*</code> constants.
   *@see asgaard.skid.RemoteBean.UPDATE_STRONG
   */
  public byte getUpdatePolicy() {
    return BeanBroker.UPDATE_STRONG;
  }
  /**
   * Get the Identifier of the Bean which shall be unique worldwide for
   * different Objects (like different Plans with luckily the same Name) and
   * common for identical Objects (like a Person reentered), <b>do not use
   * manually</b>. 
   *
   * <i>This method usually is used only once at creation time and may produce
   * different results every time.</i>
   *
   *@see asgaard.skid.SkidBean.enforceUniqueness
   *@see asgaard.skid.SkidBean.getName
   */
  public String getIdentifier() {

    StringTokenizer classTok = new StringTokenizer(this.getClass().getName(), ".");

    String className = "x";

    while (classTok.hasMoreTokens()) {

      className = classTok.nextToken();

    }

    return className+":"+String.valueOf(tsCreated);

  }

  /**
   * Returns the name of the Bean which shold be human readable and may be
   * unique in some cases, but this is not strictly required.
   */
  public String getName() {
    return "B" + String.valueOf(tsCreated);
  }

  /**
   * (Unique) Primary Key of the Bean.
   *@throws SkidException if the key is not valid now.
   */
  public int getKey() throws SkidException {
    if (key == BeanBroker.ID_FAIL)
      throw new SkidException("Key not valid now.");
    return key;
  }

  /**
    * Returns the type of the Bean's class, used for the loading-signature.
   *@throws SkidException
   */
  public int getClassKey() throws SkidException {
    if (cKey == BeanBroker.ID_FAIL)
      throw new SkidException("Key not valid now.");
    return cKey;
  }

  /**
   * Returns the type of the Bean, used for the loading-signature.
   *@throws RemoteException
   */
  public byte getBeanType() {
//  public byte getBeanType() throws RemoteException {
    return BeanBroker.TYPE_BEAN;
  }

  /**
   * Timestamp "Created" of the Bean. Based on System.currentTimeMillis().
   *@see System#currentTimeMillis
   *@throws SkidException if the timestamp is not valid now.
   */
  public long getCreated() throws SkidException {
    if (key == BeanBroker.ID_FAIL)
      throw new SkidException("Timestamp not valid now.");
    return tsCreated;
  }

  /**
   * Timestamp "Modified" of the Bean. Based on System.currentTimeMillis().
   *
   *@see System#currentTimeMillis
   *@throws SkidException if the timestamp is not valid now.
   */
  public long getModified() throws SkidException {
    if (key == BeanBroker.ID_FAIL)
      throw new SkidException("Timestamp not valid now.");
    return tsModified;
  }

  /**
   * Timestamp "Age" of the Bean (= current version by the Modified-Timestamp).
   * Based on System.currentTimeMillis().
   *
   *@see System#currentTimeMillis
   *@throws SkidException if the timestamp is not valid now.
   */
  public long getAge() throws SkidException {
    if (key == BeanBroker.ID_FAIL)
      throw new SkidException("Timestamp not valid now.");
    return tsAge;
  }

  /**
   * Sets the Age to the Modified timestamp, therefor IsChanged = false,
   * <b>do not use manually!</b> Intended usage only from RemoteBroker.
   *
   *@see asgaard.skid.RemoteBroker#put
   */
  public void setAge() {
    tsAge = tsModified;
  }
  /**
   * Returns a set of Links defined by an filterobject.
   * If this filterobject is missing (<code>null</code>) all Links of this Bean are searched.
   *@param filter the filterobject
   *@return a set of links
   *@see asgaard.skid.BeanLinkFilter
   *@throws Exception if the operation failes
   */
  public BeanLink[] getLink(BeanLinkFilter filter) throws Exception {
    BeanLinkFilter f = filter;
    if (f == null) {                                   // prepare new filter if missing...
      f = new BeanLinkFilter();
      f.setFrom( this.getKey(), BeanBroker.FILTER_EQ );
    }
    return broker.getLink( f );
  }
  /**
   * Adds a Relaition between Beans
   *@param link one relation
   *@return whether an update is possible or not.
   *@throws Exception if the key is not valid or anything else fails
   */
  public boolean putLink(BeanLink link) throws Exception {
    return broker.putLink( link );
  }
    /**
     * Adding an Event and replicating it to all suscribed Objects.
     *@param event the kond of the event (not <code>null</code>)
     */
    public void addEvent(RemoteBean event) {
      try {
        broker.addEvent(this, event);
      } catch (Exception e) {      }
    }

    /**
     * Adding an Event-Pattern and suscribing this type of event.
     *@param producer of the event (not <code>null</code>)
     *@param event the kond of the event (not <code>null</code>)
     */
    public void suscribeEvent(RemoteBean producer, RemoteBean event) {
      try {
        broker.suscribeEvent(producer, event, this);
      } catch (Exception e) {      }
    }

    /**
     * Returns the Events for this bean which are newer than the given time
     *@param newer than the given time
     *@returns a set of events
     */
    public SkidEvent[] getEvent(long newer) {
      try {
        return broker.getEvent(this, newer);
      } catch (Exception e) {
        return new SkidEvent[0];
      }
    }

  /**
   * All Keys of the Bean which are established at the first insert to the
   * RemoteBroker, <b>do not use manually</b>. Only use in <code>insertBean</code>
   *
   *@see asgaard.skid.RemoteBroker.insertBean
   *@param  pKey the generated unique primary key
   *@param cKey the key of the class
   *@param  broker  defines the owning broker
   *@return false if the key was set before.
   */
  public boolean init(int pKey, int cKey, RemoteBroker broker) {
    this.broker = broker;
    if (key == BeanBroker.ID_FAIL) {
      this.key = pKey;
      this.cKey = cKey;
      modify(null, null);
      setAge();
      tsAge = tsModified;
      return true;
    }
    return false;
  }

  /**
   * Set the modified flag and savin the old Value for the undo-mechanism (if
   * the history_policy allowes this). To avoid storing information you can use
   * the <code>history_policy</code> or use <code>null, null</code> as parameter
   * values.
   *
   *@param  oldVlaue  the old value (if any)
   *@param  fieldName the name of the field which was changed (<code>see
   * getClass.getDeclaredField()</code>)
   */
  protected void modify(Object oldValue, String fieldName) {
    if ((history_policy > BeanBroker.HISTORY_NONE)&&(oldValue != null)&&(fieldName != null))
      try {
        undoVec.addElement(new Object[] { new Long(tsModified), oldValue,
                              this.getClass().getDeclaredField(fieldName) });
        if ((history_policy > BeanBroker.HISTORY_ALL) && (undoVec.size()>history_policy)) {
          undoVec.removeElementAt( 0 );
        }
      } catch (Exception e) {
        if (debug) {
          System.err.println( e.toString());
          e.printStackTrace();
        }
      }
    tsModified = System.currentTimeMillis();
    // DIRTY TRICK: SHIFT TS 1 vvvvvvvvvvvvvvvvvvv
    if ((oldValue == this) && (fieldName == null))
      tsModified += 1;
  }

  /**
   * Has the Bean been changed?
   */
  public boolean isChanged() {
    return tsModified != tsAge;

  }

  /**
   * Ist das Bean neu?
   */
  public boolean isNew() {
    return key == BeanBroker.ID_FAIL;
  }

  /**
   * Ist the Bean read only?
   */
  public boolean isReadOnly () {

    return beanIsReadOnly || (tsCreated == 0);

  }


  /**
   * Sets (replaces if any before) the given objects.
   *@param contain the wrapped object
   *@throws SkidException if the given Object is not Serializable
   */
  public void setObject(Object contain) throws SkidException {
    if (! (contain instanceof Serializable))
      throw new SkidException("Object not Serializable.");
    modify(contain, "obj");
  }

  /**
   * Returns the wrapped Object, in this case it's the SkidBean itselfe.
   *@return this
   */
  public Object getObject() {
    return this;
  }

  /**
   * Returns a Filter for this (and all parent) class.

   */

  public SkidBeanFilter getFilter() {

    try {

      return new SkidBeanFilter( this.getClass());

    } catch (Exception e) {

      return null;

    }

  }

  /**
   * Returns an array containing Method objects reflecting all the public member

   * methods of the class or interface represented by this Class object,

   * including those declared by the class or interface and and those inherited

   * from superclasses and superinterfaces. Returns an array of length 0 if the

   * class or interface has no public member methods.

   *
   *@throws SecurityException if access to the information is denied.
   */

  public Method[] getMethods() throws SecurityException {

    return this.getClass().getMethods();

  }

  /**
   * Invokes the underlying method represented by this Method object, on the

   * specified object with the specified parameters. Individual parameters are

   * automatically unwrapped to match primitive formal parameters, and both
   * primitive and reference parameters are subject to widening conversions as
   * necessary. The value returned by the underlying method is automatically
   * wrapped in an object if it has a primitive type.
   *
   *@param name the name of the Method
   *@param args the Arguments
   *@param modify call the modify-method too? (not including undo)
   *@return a result returned (if any)
   *@see java.lang.reflect.Method#invoke
   *@throws IllegalAccessException if the underlying method is inaccessible.

   *@throws IllegalArgumentException if the number of actual and formal
   * parameters differ, or if an unwrapping conversion fails.
   *@throws InvocationTargetException if the underlying method throws an
   * exception.
   *@throws NullPointerException if the specified object is null.
   *@throws NoSuchMethodException if there is no Method with this name
   */
  public Object invoke(String name,Object args[], boolean modify) throws IllegalAccessException,

                                                    IllegalArgumentException,

                                                    InvocationTargetException,

                                                    NullPointerException,

                                                    NoSuchMethodException {

    if (args == null)

      args = new Object[0];

    Class[] params = new Class[args.length];

    for (int i=0;i<args.length;i++) {

      params[i] = args[i].getClass();

    }

    Object ret = this.getClass().getMethod(name, params).invoke(this, args);

    if (modify)

      modify(null, null);

    return ret;

  }

  /**
   * Save all modifications (if any) and checks if the update is possible
   * (timestamp of the RemoteBroker). Remove the modified flag, set the
   * modified-timestamp.
   *
   *@throws Exception if the save failed.
   */
  public void flush() throws Exception {
    if (broker == null)
      broker = BeanBroker.THIS;
    if (broker == null)
      throw new SkidException("Bean not bound to any broker.");
    if (debug) broker.logMsg( this.getClass().getName(), "flush#0", LogMsg.LOG_DEBUG, true, "CHANGED: "+String.valueOf(isChanged()));
    if (isChanged()) {
      Object[] keys = broker.putBean(this);
      if (debug) broker.logMsg( this.getClass().getName(), "flush#1", LogMsg.LOG_DEBUG, keys!=null, String.valueOf(keys));
      if (keys==null) {
        throw new SkidException("Newer Bean exists");
      }
      key = ((Integer)keys[0]).intValue();
      cKey = ((Integer)keys[1]).intValue();
      tsModified = ((Long)keys[2]).longValue();
      setAge();
      undoVec.removeAllElements();
      if (debug) broker.logMsg( this.getClass().getName(), "flush#2", LogMsg.LOG_DEBUG, true, "AGE: "+String.valueOf(getAge()));
    }
  }

  /**
   * Clones the Object and does all the settings neccesary to be assigned as a

   * new SkidBean. To make it persistent use the SkidBean.flush() or the

   * RemoteBroker.put() method. The call of these methods resets the

   * UNDO-mechanism and clear all relied information.

   *

   *@see asgaard.skid.RemoteBean#flush

   *@see asgaard.skid.RemoteBroker#put

   */

  public Object cloneBean() throws CloneNotSupportedException {

//  public Object cloneBean() throws CloneNotSupportedException, RemoteException {

    RemoteBean newBean = null;

    int mykey = key;

    int mycKey = cKey;

    long myCreated = tsCreated;
    long myModified = tsModified;
    long myAge = tsAge;
    boolean myReadOnly = beanIsReadOnly;
    try {
      key = BeanBroker.ID_FAIL;

      cKey = mycKey;

      tsCreated = System.currentTimeMillis()-1;
      tsModified = tsCreated+1;
      tsAge = tsCreated;
      beanIsReadOnly = false;
	    newBean = (RemoteBean) super.clone();
    } finally {

      key = mykey;

      cKey = mycKey;

      tsCreated = myCreated;
      tsModified = myModified;
      tsAge = myAge;
      beanIsReadOnly = myReadOnly;
    }
    return newBean;
  }

  /**
   * Produces a discription of the last changes which can be undone.
   *@return a set of description-strings
   */
  public String[] getChangeHistory() {
    String[] hist = new String[undoVec.size()];
    for (int i=0;i<hist.length;i++) {
      Object[] h = (Object[]) undoVec.elementAt(i);
      if (h[2] != null) { // field
        Field f = (Field) h[2];
        hist[i] = f.getName() + ": ";
        if (h[1] != null) { // old
          hist[i] += h[1].toString();
        }
      }
    }
    return hist;
  }

  /**
   * Undo the last <code>steps</code> changes (<code>0</code> = all,

   * <code>-1</code> no undo but ask if any undos possible) <b>HAS TO BE THE

   * LAST METHOD OF A REMOTEBEAN ROOT-IMPLEMENTATION</b>.

   * The call of the <code>flush()</code> methods resets the UNDO-mechanism and

   * clear all relied information.

   *<p>

   * <i>Remark:</i> Undo only works if the <code>modify()</code> method was

   * called by the <code>setXXXX</code> method before change.

   *

   *@see asgaard.skid.RemoteBean.HISTORY_ALL

   *@see asgaard.skid.RemoteBean.HISTORY_NONE
   *@see modify
   *@param steps cont of changes to undo

   *@return if there are steps left which can be undone

   */

  public boolean undo(int steps) {

    if (steps < 0)                                // calc position in history

      return (undoVec.size() > 0);

    if (undoVec.size() == 0)

      return false;

    int i = steps;

    if (steps == 0)

      i = undoVec.size();

    try {                                         // get History-object

      while ((i>0) && (undoVec.size()>0))

        try {

          Object[] h = (Object[]) undoVec.lastElement();

          Long ts = (Long) h[0];

          tsModified = ts.longValue();            // undo it

          if ((h[1] != null) && (h[2] != null)) {

            Field f = (Field) h[2];

            f.set(this, h[1]);
          }
        } finally {

          i--;

          undoVec.removeElementAt(undoVec.size()-1);

        }

    } catch (Exception e) {

      if (debug) {

        System.err.println( e.toString());
        e.printStackTrace();
      }
    }

    return (undoVec.size() > 0);

  }


}



