// This class is part of Skid
//@autor  (P) by Klaus Hammermller, mailto:klaus@ifs.tuwien.ac.at
//@version	1.0 130499
//
//@created	130499  KH
package asgaard.skid;

// Import the java classes used
import java.util.*;
import java.io.*;

import java.lang.reflect.*;

/**
 * Implementation of an wrapper for a Streamable Object carried by an DataBeanContainer,
 * including the functionality (persistence) of an SkidBean.
 *<p>
 *<b><i>Important:</i></b> Key values which are accessible via the <code>RemoteBroker</code>
 * have to be defined following the <i>Java-Beans-Convention</i> using the <i>getXXXX</i>
 * methods for key-generation.
 *<p>
 * <b><i>Remark:</i></b> All attribues used by the Object which
 * are not intended to be persitent have to be declared <code>transient</code> and restored
 * by default or the parameterless constructor.
 *<p>
 *<b><i>Minimal Usage:</i></b> (The RemoteBroker should be instanciated at this time)
 *<code>  SkidBean b = new SkidBean( <object> );</code><br>
 *<code>  b.setName() </code> (optional) <br>
 *<code>  b.flush();</code><br>
 *
 *@see asgaard.skid.SkidBean
 */

public class SkidBeanContainer extends SkidBean {

  private Object obj = null;
  /**
   * Creates a Conainer wrapping a given object.
   *@param contain the wrapped object
   *@throws SkidException if the given Object is not Serializable
   */
  public SkidBeanContainer(Object contain) throws Exception {
    super ();
    setObject(contain);
  }

  /**
   * Returns the name of the Bean which shold be human readable and may be
   * unique in some cases, but this is not strictly required.
   */
  public String getName() {
    return String.valueOf(obj);
  }

  /**
   * Sets (replaces if any before) the given objects.
   *@param contain the wrapped object
   *@throws SkidException if the given Object is not Serializable
   */
  public void setObject(Object contain) throws SkidException {
    if (! (contain instanceof Serializable))
      throw new SkidException("Object not Serializable.");
    modify(contain, "obj");
  }

  /**
   * Returns the wrapped Object
   *@return the object, if any.
   */
  public Object getObject() {
    return obj;
  }

  /**
   * Returns an array containing Method objects reflecting all the public member methods of the class or interface represented by this Class object, including those

   * declared by the wrapped object.

   *
   *@throws SecurityException if access to the information is denied.
   */

  public Method[] getMethods() throws SecurityException {

    Method[] container = this.getClass().getMethods();

    if (obj == null)

      return container;

    Method[] wrapped = obj.getClass().getMethods();

    Method[] all = new Method[container.length + wrapped.length];

    for (int i=0;i<container.length;i++) {

      all[i] = container[i];

    }

    for (int j=0;j<wrapped.length;j++) {

      all[j+container.length] = wrapped[j];

    }

    return all;

  }


  /**
   * Invokes the underlying method represented by this Method object, on the specified object with the specified parameters. Individual parameters are

   * automatically unwrapped to match primitive formal parameters, and both primitive and reference parameters are subject to widening conversions as
   * necessary. The value returned by the underlying method is automatically wrapped in an object if it has a primitive type.
   *<p>
   * Search order: First methods of the <code>SkidBeanContainer</code> are invoked before the Methods of the wrapped Object are invoked.
   *
   *@param name the name of the Method
   *@param args the Arguments
   *@param modify call the modify-method too? (not including undo)
   *@return a result returned (if any)
   *@see java.lang.reflect.Method#invoke
   *@throws IllegalAccessException if the underlying method is inaccessible.

   *@throws IllegalArgumentException if the number of actual and formal parameters differ, or if an unwrapping conversion fails.
   *@throws InvocationTargetException if the underlying method throws an exception.
   *@throws NullPointerException if the specified object is null.
   *@throws NoSuchMethodException if there is no Method with this name
   */
  public Object invoke(String name, Object args[], boolean modify) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, NullPointerException, NoSuchMethodException {

    Class[] params = new Class[args.length];

    for (int i=0;i<args.length;i++) {

      params[i] = args[i].getClass();

    }

    Object ret = null;

    try {

      ret = this.getClass().getMethod(name, params).invoke(this, args);

      if (modify)

        modify(null, null);

      return ret;

    } catch (NoSuchMethodException e) {

      if (obj == null)

        throw e;

      ret = obj.getClass().getMethod(name, params).invoke(obj, args);

      if (modify)

        modify(null, null);

      return ret;

    }

  }

}