// This class is part of Skid
//@autor  (P) by Klaus Hammermller, mailto:klaus@ifs.tuwien.ac.at
//@version	2.0 080499
//
//@created	080499  KH interface to class SkidBean
package asgaard.skid;

// Import the java classes used
import java.rmi.*;

import java.lang.reflect.*;

/**
 * Interface of an Data-Bean, normaly use <b><code>SkidBean</code></b> where all nessacary
 * methods are implemented. This interface is a workaround if polymorpy is nessecary and to
 * implement the Remote-Interface automatically. Therefor all Beans are accessible via RMI.
 *<p>
 *<b><i>Important:</i></b> Key values which are accessible via the <code>BeanBroker</code>
 * have to be defined following the <i>Java-Beans-Convention</i> using the <i>getXXXX</i>
 * methods for key-generation.
 *<p>
 * <b><i>Remark:</i></b> All attribues used by the Object which
 * are not intended to be persitent have to be declared <code>transient</code> and restored
 * by default or the parameterless constructor.
 *<p>
 * <i>Remark:</i> A System-Object cannot be modified and carries the created-Timestamp 0
 *
 *@see asgaard.skid.SkidBean
 *@see asgaard.skid.BeanBroker
 */
public interface SkidBeanInterface extends Remote {

  /**
   * Get Link to the BeanBroker (one per VM).
   */
  public BeanBroker getBroker();
  /**
   * Get's the update policy defined by the <code>UPDATE_*</code> constants.
   *@see asgaard.skid.SkidBeanInterface.UPDATE_ALL
   *@see asgaard.skid.SkidBeanInterface.UPDATE_WEAK
   *@see asgaard.skid.SkidBeanInterface.UPDATE_STRONG
   */
  public byte getUpdatePolicy();
  /**
   * Get the Identifier of the Bean which shall be unique worldwide for different
   * Objects (like different Plans with luckily the same Name) and common for
   * identical Objects (like a Person reentered), <b>do not use manually</b>.
   */
  public String getIdentifier();

  /**
   * Returns the name of the Bean which shold be human readable and may be unique in some cases,
   * but this is not strictly required.
   */
  public String getName();

  /**
   * (Unique) Primary Key of the Bean.
   *@throws SkidException if the key is not valid now.
   */
  public int getKey() throws SkidException;

  /**
   * Timestamp "Created" of the Bean. Based on System.currentTimeMillis().
   *@see System#currentTimeMillis
   *@throws SkidException if the timestamp is not valid now.
   */
  public long getCreated() throws SkidException;

  /**
   * Timestamp "Modified" of the Bean. Based on System.currentTimeMillis().
   *@see System#currentTimeMillis
   *@throws SkidException if the timestamp is not valid now.
   */
  public long getModified() throws SkidException;

  /**
   * Timestamp "Age" of the Bean (= current version by the Modified-Timestamp). Based on System.currentTimeMillis().
   *@see System#currentTimeMillis
   *@throws SkidException if the timestamp is not valid now.
   */
  public long getAge() throws SkidException;

  /**
   * Sets the Age to the Modified timestamp, therefor IsChanged = false, <b>do not use manually!</b>
   * Intended usage only from BeanBroker.
   *@see asgaard.skid.BeanBroker#put
   */
  public void setAge();
  /**
   * Returns a set of Links defined by an filterobject.
   * If this filterobject is missing (<code>null</code>) all Links of this Bean are searched.
   *@param filter the filterobject
   *@return a set of links
   *@see asgaard.skid.BeanLinkFilter
   *@throws Exception if the search failes
   */
  public BeanLink[] getLink(BeanLinkFilter filter) throws Exception;
  /**
   * Adds a Relaition between Beans
   *@param link one relation
   *@return wether an update is possible or not.
   *@throws Exception if the key is not valid or anything else fails
   */
  public boolean putLink(BeanLink link) throws Exception;

    /**
     * Adding an Event and replicating it to all suscribed Objects.
     *@param event the kond of the event (not <code>null</code>)
     */
    public void addEvent(SkidBeanInterface event);
    /**
     * Adding an Event-Pattern and suscribing this type of event.
     *@param producer of the event (not <code>null</code>)
     *@param event the kond of the event (not <code>null</code>)
     */
    public void suscribeEvent(SkidBeanInterface producer, SkidBeanInterface event);
    /**
     * Returns the Events for this bean which are newer than the given time
     *@param newer than the given time
     *@returns a set of events
     */
    public SkidEvent[] getEvent(long newer);
  /**
   * All Keys of the Bean which are established at the first insert to the BeanBroker,
   *<b>do not use manually</b>. Only use in <code>insertBean</code>.
   *
   *@see asgaard.skid.BeanBroker.insertBean
   *@param  pKey the generated unique primary key
   *@param  broker  defines the owning broker
   *@return false if the key was set before.
   */
  public boolean init(int pKey, BeanBroker broker);

  /**
   * Has the Bean been changed?
   */
  public boolean isChanged();
  /**
   * Ist das Bean neu?
   */
  public boolean isNew();
  /**
   * Ist the Bean read only?
   */
  public boolean isReadOnly ();


  /**

   * Returns the wrapped Object (if wrapped by a container) or the Bean (if it's no wrapper)
   *@return the object, if any.
   */
  public Object getObject();


  /**

   * Returns a Filter for this (and all parent) class.

   */

  public SkidBeanFilter getFilter();


  /**

   * Returns an array containing Method objects reflecting all the public member methods of the class or interface represented by this Class object, including those

   * declared by the class or interface and and those inherited from superclasses and superinterfaces. Returns an array of length 0 if the class or interface has no
   * public member methods.
   *
   *@see java.lang.reflect.Method#getMethods
   *@throws SecurityException if access to the information is denied.
   */

  public Method[] getMethods() throws SecurityException;


  /**

   * Invokes the underlying method represented by this Method object, on the specified object with the specified parameters. Individual parameters are

   * automatically unwrapped to match primitive formal parameters, and both primitive and reference parameters are subject to widening conversions as
   * necessary. The value returned by the underlying method is automatically wrapped in an object if it has a primitive type.
   *
   *@param name the name of the Method
   *@param args the Arguments
   *@param modify call the modify-method too? (not including undo)
   *@return a result returned (if any)
   *@see java.lang.reflect.Method#invoke
   *@throws IllegalAccessException if the underlying method is inaccessible.

   *@throws IllegalArgumentException if the number of actual and formal parameters differ, or if an unwrapping conversion fails.
   *@throws InvocationTargetException if the underlying method throws an exception.
   *@throws: NullPointerException if the specified object is null.
   *@throws NoSuchMethodException if there is no Method with this name
   */
  public Object invoke(String name, Object args[], boolean modify) throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, NullPointerException, NoSuchMethodException;

  /**
   * Save all modifications (if any) and checks if the update is possible (timestamp of the BeanBroker).
   * Remove the modified flag, set the modified-timestamp. The call of this methods resets
   * the UNDO-mechanism and clear all relied information.
   *
   *@throws Exception if the save failed.
   */
  public void flush() throws Exception;

  /**
   * Produces a discription of the last changes which can be undone.
   *@return a set of description-strings
   */
  public String[] getChangeHistory();
  /**
   * Undo the last <code>steps</code> changes (<code>0</code> = all,

   * <code>-1</code> no undo but ask if any undos possible). The call of the

   * <code>flush()</code> methods resets the UNDO-mechanism and clear all relied information.

   *

   *@param steps cont of changes to undo

   *@return if there are steps left which can be undone

   */

  public boolean undo(int steps);

  /**
   * Clones the Object and does all the settings neccesary to be assigned as a new

   * SkidBean. To make it persistent use the SkidBean.flush() or the BeanBroker.put()

   * method.

   *@see asgaard.skid.SkidBeanInterface#flush

   *@see asgaard.skid.BeanBroker#put

   */

  public Object clone() throws CloneNotSupportedException;


}
