package asgaard.utils.cryptography;

/**
 *	<P>Class Cryptography is the abstract class that contains abstract
 *	forms of all cryptography functions (the methods here will just
 *	throw no such method exceptions).  Individual subclasses will
 *	implement different cryptographic algorithms, but application
 *	programmers should primarily program against this interface.</P>
 *
 *	<P>Copyright &copy; 1997 David Purdue.</P>
 *
 *	@author David Purdue <David.Purdue@Aus.Sun.COM>
 *	@version 1.6 - 97/05/28 11:54:11
 */

public abstract class Cryptography
{
	/*
	 *	The name of the Java Cryptography package.
	 */
	private final static String CRYPTO_PACAKAGE_NAME = "asgaard.utils.cryptography";

	/**
	 *	The getCrypto method takes the name of a cryptographic
	 *	algorithm and generates an object that implements it.
	 *
	 *	@param cryptoName The name of the algorithm to use.
	 *	@return An object of a subclass of Cryptography - a cryptosystem.
	 *	@exception asgaard.utils.cryptography.NoSuchCryptoException if the name specified is not a sublass of Cryptography.
	 */
	public static Cryptography getCrypto(String cryptoName)
						throws NoSuchCryptoException
	{
		Class cryptoClass;
		Object newInstance;

		// Make sure that we choose a class from the Crypto package.
		String classname = CRYPTO_PACAKAGE_NAME + "." + cryptoName;

		try
		{
			cryptoClass = Class.forName(classname);
		}
		catch (ClassNotFoundException e)
		{
			// Oops, no such class!
			String gripe = "No implementation for the "
					+ cryptoName + " cryptosystem.";
			throw new NoSuchCryptoException(gripe);
		}

		// Now check that the class is a cryptosystem, i.e. is a
		// subclass of Cryptography.  Work up the inheiritance tree,
		// if we get to Cryptography, good.  If we hit a null pointer,
		// then we have run out of classes, so bad.
		Class tmpClass = cryptoClass;
		while (!(tmpClass.getName().equals(CRYPTO_PACAKAGE_NAME
							+ ".Cryptography")))
		{
			tmpClass = tmpClass.getSuperclass();
			if (tmpClass == null)
			{
				String gripe = cryptoName
						+ " is not a cryptosystem.";
				throw new NoSuchCryptoException(gripe);
			}
		}

		// OK, we are happy with the class, let's try to make an object.
		try
		{
			newInstance = cryptoClass.newInstance();
		}
		catch (InstantiationException e)
		{
			String gripe = "Cannot create instances of the "
						+ cryptoName + " class.";
			throw new NoSuchCryptoException(gripe);
		}
		catch (IllegalAccessException e)
		{
			String gripe = "Illegal Access for " + cryptoName
								+ " class.";
			throw new NoSuchCryptoException(gripe);
		}

		return (Cryptography) newInstance;
	} // getCrypto()



	/**
	 *	Place holder for encryption function.
	 *
	 *	@param plaintext The plain text to encrypt as an array of bytes.
	 *	@return An array of bytes with the cipher text.
	 *	@exception java.lang.NoSuchMethodException Thrown if a subclass does not implement this function and a client tries to use it anyway.
	 */
	public byte[] encrypt(byte[] plaintext) throws NoSuchMethodException
	{
		String gripe;

		gripe = "Class " + this.getClass().getName() +
			" does not implement encrypt().";
		throw new NoSuchMethodException(gripe);
	} // encrypt()



	/**
	 *	Place holder for encryption tidy up.
	 *
	 *	@return An array of bytes with the cipher text.
	 *	@exception java.lang.NoSuchMethodException Thrown if a subclass does not implement this function and a client tries to use it anyway.
	 */
	public byte[] encryptFinish() throws NoSuchMethodException
	{
		String gripe;

		gripe = "Class " + this.getClass().getName() +
			" does not implement encryptFinish().";
		throw new NoSuchMethodException(gripe);
	} // encryptFinish()



	/**
	 *	Place holder for decryption.
	 *
	 *	@param ciphertext The cipher text to decrypt, an array of bytes.
	 *	@return An array of bytes with the plain text.
	 *	@exception java.lang.NoSuchMethodException Thrown if a subclass does not implement this function and a client tries to use it anyway.
	 */
	public byte[] decrypt(byte[] ciphertext) throws NoSuchMethodException
	{
		String gripe;

		gripe = "Class " + this.getClass().getName() +
			" does not implement decrypt().";
		throw new NoSuchMethodException(gripe);
	} // decrypt()



	/**
	 *	Place holder for decryption tidy up.
	 *
	 *	@return An array of bytes with the plain text.
	 *	@exception java.lang.NoSuchMethodException Thrown if a subclass does not implement this function and a client tries to use it anyway.
	 */
	public byte[] decryptFinish() throws NoSuchMethodException
	{
		String gripe;

		gripe = "Class " + this.getClass().getName() +
			" does not implement decryptFinish().";
		throw new NoSuchMethodException(gripe);
	} // decryptFinish()



	/**
	 *	Place holder for setting a key.
	 *
	 *	@param key an array of bytes containing the key.
	 *	@exception java.lang.NoSuchMethodException Thrown if a subclass does not implement this function and a client tries to use it anyway.
	 */
	public void setKey(byte[] key) throws NoSuchMethodException
	{
		String gripe;

		gripe = "Class " + this.getClass().getName() +
			" does not implement setKey().";
		throw new NoSuchMethodException(gripe);
	} // decryptFinish()



	/**
	 *	Place holder for getting a key.
	 *
	 *	@return An array of bytes with the user key.
	 *	@exception java.lang.NoSuchMethodException Thrown if a subclass does not implement this function and a client tries to use it anyway.
	 */
	public byte[] getKey() throws NoSuchMethodException
	{
		String gripe;

		gripe = "Class " + this.getClass().getName() +
			" does not implement getKey().";
		throw new NoSuchMethodException(gripe);
	} // decryptFinish()
} // class Cryptography
