package asgaard.utils.cryptography;

/**
 *	<P>IDEA.java - A java implementation of the IDEA block cipher.
 *	IDEA (International Data Encryption Algorithm) developed by
 *	Xuejia Lai and James L. Massey of ETH Zurich.</P>
 *
 *	<P>NOTE: The IDEA(tm) block cipher is covered by patents held by ETH
 *	and Ascom-Tech AG, Swiss patent number PCT/CH91/00117, European
 *	patent number EP 0 482 154 B1, U.S. patent number US005214703.
 *	IDEA(tm) is a trademark of Ascom-Tech AG.  There is no license
 *	fee required for noncommercial use.  Commercial users may obtain
 *	licensing details from Dieter Profos, Ascom Tech AG, Solothurn Lab,
 *	Postfach 151, 4502 Solothurn, Switzerland, Tel +41 65 242885,
 *	Fax +41 65 235761.</P>
 *
 *	<P>IDEA uses a lot of operations on unsigned 16 bit integers.  Since
 *	Java has no unsigned integers these will be stored and manipulated
 *	as 32 bit integers.</P>
 *
 *	<P>Copyright &copy; 1997 David Purdue.</P>
 *
 *	@author David Purdue <David.Purdue@Aus.Sun.COM>
 *	@version 1.2 - 97/05/28 11:52:11
 *	@see asgaard.utils.cryptography.Cryptography
 */

public class IDEA extends Cryptography
{
	// How many bytes in an IDEA user key?
	private static final int KEYSIZE = 16;

	// How many bytes in an IDEA encryption block?
	private static final int BLOCKSIZE = 8;

	// How many rounds to run the cipher?
	private static final int ROUNDS = 8;

	// How long is an expanded IDEA encryption/decryption key?
	private static final int KEYLEN = (6 * ROUNDS) + 4;



	// Store the user key and the exanded encryption and decryption keys.
	private byte[] userKey = new byte[KEYSIZE];
	private int[] encryptKey = new int[KEYLEN];
	private int[] decryptKey = new int[KEYLEN];



	/**
	 *	Constructor - set up IDEA object with default key (all
	 *	zeros).
	 */
	public IDEA()
	{
		for (int i=0; i < KEYSIZE; i++)
		{
			userKey[i] = (byte) 0;
		}

		// Set up expanded encrypt/decrypt keys.
		expandKey();
		invertKey();
	}



	/*
	 *	unsigned() - in Java, bytes are always a signed quantity
	 *	(i.e. values -128 - 127).  We want to deal with the unsigned
	 *	value (0 - 255).  This method takes a byte and returns an
	 *	integer in the range 0-255.
	 */
	private static int unsigned(byte b)
	{
		// if (b >= 0)
		// {
		// 	return (int)b;
		// }
		// 
		// return 0x80 + (b & 0x7f);

		return (b >= 0) ? ((int) b) : ((int)(0x100 + b));
	} // unsigned()

	/*
	 *	IDEA is built around three basic operations: multiplication
	 *	modulo (2**16+1), addition modulo (2**16) and XOR.  The
	 *	first two are implemented in mul() and add().
	 */

//	/*
//	 *	mul() - multiply mod 2**16+1 = 65537.  This code adapted
//	 *	from Xuejia Lai's C implementation.  This looks strange
//	 *	because 2**16 (0x10000) is represented in 16 bits by 0x0000.
//	 */
//	private static int mul(int a, int b)
//	{
//		// cast to longs to keep things positive.
//		long p = (((long) a) * ((long b))) & 0xffff;
//
//		if (p != 0)
//		{
//			int t1 = (int)(p & 0xffff);
//			int t2 = (int)(p >>> 16);
//
//			if (t1 < t2)
//			{
//				return 0x10001 + t1 - t2;
//			}
//			else
//			{
//				return t1 - t2;
//			}
//			// return (t1 - t2) + ((t1 < t2)? 1 : 0);
//		}
//		else if (a != 0)
//		{
//			return (0x10001 - a) & 0xffff;
//		}
//		else
//		{
//			return (0x10001 - b) & 0xffff;
//		}
//	}

	/*
	 *	mul() - multiply mod 2**16+1 = 0x10001 = 65537.  0x0000
	 *	represents 0x10000.
	 */
	private static int mul(int a, int b)
	{
		if (a == 0) a = 0x10000;
		if (b == 0) b = 0x10000;
		return (int)(((((long) a) * ((long) b)) % 0x10001) & 0xffff);
	} // mul()


	/*
	 *	add() - modulo 0x10000.
	 */
	private static int add(int a, int b)
	{
		return (a + b) & 0xffff;
	} // add()



	/*
	 *	mulInv() - work out the multiplicative inverse of
	 *	a number, where mul is defined above. (i.e. for x,
	 *	calculate y such that mul(x, y) == 1. Uses Euclid's
	 *	algorithm.
	 */
	private static int mulInv(int x)
	{
		int t0, t1, y, q;

		if (x <= 1)
		{
			// 0 and 1 are their own inverses.
			return x;
		}

		t1 = 0x10001 / x;
		y = 0x10001 % x;
		// remember, 0 < x <= 0xffff, hence 0 <= y <= 0xfffe;

		if (y == 1)
		{
			return (0x10001 - t1) & 0xffff;
		}

		t0 = 1;
		do
		{
			q = x / y;
			x = x % y;
			t0 += q * t1;
			if (x == 1)
			{
				return t0;
			}
			q = y / x;
			y = y % x;
			t1 += q * t0;
		} while (y != 1);
		return (0x10001 - t1) & 0xffff;
	} // mulInv()



	/*
	 *	expandKey() - work out the expanded encryption key from
	 *	the 128-bit user key.
	 */
	private void expandKey()
	{
		// First 8 keys are taken directly from user key.  Take
		// pairs of bytes as 16 bit words.
		for (int i = 0; i < 8; i++)
		{
			encryptKey[i] = (unsigned(userKey[2*i]) << 8) | 
					(unsigned(userKey[(2*i)+1]));
		}

		// Subsequent keys are done by rotation of original key.
		// At this stage, is is simpler to translate Lai's code
		// than try to get the bit manipulation right in a more
		// Java'ish manner - that will come later.
		int EKstart = 0;
		int i = 0;
		for (int j = 8; j < KEYLEN; j++)
		{
			i++;
			encryptKey[EKstart + i + 7] =
				(
				  (encryptKey[EKstart + (i & 7)] << 9) |
				  (encryptKey[EKstart + ((i+1) & 7)] >>> 7)
				) & 0xffff;
			EKstart += i & 8;
			i &= 7;
		}
	} // expandKey()



	/*
	 *	invertKey() - work out the decryption key from the encryption
	 *	key.
	 */
	private void invertKey()
	{
		// Done mostly by translating Lai's code.  Lai worked on 16 bit
		// unsigned ints - which is the basis for negation.
		int EKptr = 0;
		int DKptr = KEYLEN;

		decryptKey[--DKptr] = mulInv(encryptKey[EKptr + 3]);
		decryptKey[--DKptr] = (0x10000 - encryptKey[EKptr + 2]) & 0xffff;
		decryptKey[--DKptr] = (0x10000 - encryptKey[EKptr + 1]) & 0xffff;
		decryptKey[--DKptr] = mulInv(encryptKey[EKptr]);
		EKptr += 4;

		for (int i = 0; i < (ROUNDS - 1); i++)
		{
			decryptKey[--DKptr] = encryptKey[EKptr + 1];
			decryptKey[--DKptr] = encryptKey[EKptr];
			EKptr += 2;

			decryptKey[--DKptr] = mulInv(encryptKey[EKptr + 3]);
			decryptKey[--DKptr] = (0x10000 - encryptKey[EKptr + 1]) & 0xffff;
			decryptKey[--DKptr] = (0x10000 - encryptKey[EKptr + 2]) & 0xffff;
			decryptKey[--DKptr] = mulInv(encryptKey[EKptr]);
			EKptr += 4;
		}

		decryptKey[--DKptr] = encryptKey[EKptr + 1];
		decryptKey[--DKptr] = encryptKey[EKptr];
		EKptr += 2;

		decryptKey[--DKptr] = mulInv(encryptKey[EKptr + 3]);
		decryptKey[--DKptr] = (0x10000 - encryptKey[EKptr + 2]) & 0xffff;
		decryptKey[--DKptr] = (0x10000 - encryptKey[EKptr + 1]) & 0xffff;
		decryptKey[--DKptr] = mulInv(encryptKey[EKptr]);
	} // invertKey()



	/**
	 *	ideaCipher() - encrypt or decrypt a 64 bit block with the
	 *	IDEA cipher.
	 */
	private static byte[] ideaCipher(byte[] text, int[] key)
	{
		int	x1, x2, x3, x4, s2, s3;
		int	keyPtr = 0;

		x1 = (unsigned(text[0]) << 8) + unsigned(text[1]);
		x2 = (unsigned(text[2]) << 8) + unsigned(text[3]);
		x3 = (unsigned(text[4]) << 8) + unsigned(text[5]);
		x4 = (unsigned(text[6]) << 8) + unsigned(text[7]);

		for (int r = 0; r < ROUNDS; r++)
		{
			x1 = mul(x1, key[keyPtr++]);
			x2 = add(x2, key[keyPtr++]);
			x3 = add(x3, key[keyPtr++]);
			x4 = mul(x4, key[keyPtr++]);

			s3 = x3;
			x3 = x3 ^ x1;
			x3 = mul(x3, key[keyPtr++]);
			s2 = x2;
			x2 = x2 ^ x4;
			x2 = add(x2, x3);
			x2 = mul(x2, key[keyPtr++]);
			x3 = add(x3, x2);

			x1 = x1 ^ x2;
			x4 = x4 ^ x3;

			x2 = x2 ^ s3;
			x3 = x3 ^ s2;
		} // idea cipher round.

		x1 = mul(x1, key[keyPtr++]);
		x3 = add(x3, key[keyPtr++]);
		x2 = add(x2, key[keyPtr++]);
		x4 = mul(x4, key[keyPtr]);

		byte[] outBuf = new byte[BLOCKSIZE];

		outBuf[0] = (byte)((x1 >> 8) & 0xff);
		outBuf[1] = (byte)(x1 & 0xff);
		outBuf[2] = (byte)((x3 >> 8) & 0xff);
		outBuf[3] = (byte)(x3 & 0xff);
		outBuf[4] = (byte)((x2 >> 8) & 0xff);
		outBuf[5] = (byte)(x2 & 0xff);
		outBuf[6] = (byte)((x4 >> 8) & 0xff);
		outBuf[7] = (byte)(x4 & 0xff);

		return outBuf;
	} // ideaCipher()



	/*
	 *	ecbEn() performs IDEA encryption in Electronic Code Book
	 *	mode. ecbEnFinish() tidies up - padding and encrypting the
	 *	last block.
	 */
	private byte[] encryptBuffer = new byte[BLOCKSIZE];
	private int encryptBufferPtr = 0;

	private byte[] ecbEn(byte[] plain)
	{
		byte[] cipher;
		int plainPtr = 0;
		int cipherPtr = 0;

		// Work out how many bytes of ciphertext we will return.
		// We return whole blocks, and the number will be the
		// number of bytes already in the buffer, plus the number
		// of bytes given here, divided by block size (remainder
		// is left in the buffer for next time or finish).
		int numblocks = (encryptBufferPtr + plain.length) / BLOCKSIZE;
		cipher = new byte[BLOCKSIZE * numblocks];

		// Effectively, we deal with plain text a byte at a time.
		while (plainPtr < plain.length)
		{
			// Try to fill the encrypt buffer.
			int room = Math.min(
					(BLOCKSIZE - encryptBufferPtr),
					(plain.length - plainPtr)
					   );
			System.arraycopy(plain, plainPtr,
					 encryptBuffer, encryptBufferPtr,
					 room);
			plainPtr += room;
			encryptBufferPtr += room;

			// If we filled a block, then encrypt it.
			if (encryptBufferPtr == BLOCKSIZE)
			{
				byte[] cipherText =
					ideaCipher(encryptBuffer, encryptKey);
				System.arraycopy(cipherText, 0,
						 cipher, cipherPtr,
						 BLOCKSIZE);

				cipherPtr += BLOCKSIZE;
				encryptBufferPtr = 0;
			}
		}

		return cipher;
	} // ecbEn()

	private byte[] ecbEnFinish()
	{
		// Store number of bytes of padding in to last byte.
		encryptBuffer[BLOCKSIZE - 1] = 
				(byte)(BLOCKSIZE - encryptBufferPtr);

		// Fill the rest with zeros.
		for ( ; encryptBufferPtr < (BLOCKSIZE - 1); encryptBufferPtr++)
		{
			encryptBuffer[encryptBufferPtr] = (byte) 0;
		}

		// tidy up for next use.
		encryptBufferPtr = 0;

		// Encrypt last block and return.
		return ideaCipher(encryptBuffer, encryptKey);
	} // ecbEnFinish()



	/*
	 *	ecbDe() performs IDEA decryption in Electronic Code Book
	 *	mode. ecbDeFinish() tidies up - Removing the padding from
	 *	the last block.
	 */
	private byte[] decryptBuffer = new byte[BLOCKSIZE];
	private int decryptBufferPtr = 0;
	private byte[] lastBlock = null;

	private byte[] ecbDe(byte[] cipher)
	{
		byte[] plain;
		int plainPtr = 0;
		int cipherPtr = 0;

		// Work out how many bytes of plaintext we will return.
		// We return whole blocks, and the number will be the
		// number of bytes already in the buffer, plus the number
		// of bytes given here, divided by block size (remainder
		// is left in the buffer for next time or finish).  Subtract
		// one for the block we remember if we do not have one
		// remembered.
		int numblocks = (decryptBufferPtr + cipher.length) / BLOCKSIZE;
		if (lastBlock == null) numblocks--;
		plain = new byte[BLOCKSIZE * numblocks];

		// Effectively, we deal with cipher text a byte at a time.
		while (cipherPtr < cipher.length)
		{
			// Try to fill the decrypt buffer.
			int room = Math.min(
					(BLOCKSIZE - decryptBufferPtr),
					(cipher.length - cipherPtr)
					   );
			System.arraycopy(cipher, cipherPtr,
					 decryptBuffer, decryptBufferPtr,
					 room);
			cipherPtr += room;
			decryptBufferPtr += room;

			// If we filled a block, then decrypt it.
			if (decryptBufferPtr == BLOCKSIZE)
			{
				if (lastBlock != null)
				{
					System.arraycopy(lastBlock, 0,
							 plain, plainPtr,
							 BLOCKSIZE);
					plainPtr += BLOCKSIZE;
				}
				lastBlock =
					ideaCipher(decryptBuffer, decryptKey);

				decryptBufferPtr = 0;
			}
		}

		return plain;
	} // ecbDe()

	private byte[] ecbDeFinish()
	{
		// if decryptBufferPtr != 0, then we were not given whole
		// blocks of ciphertext to decrypt.  Last block is probably
		// garbage anyway, so do nothing - should we have a mechanism
		// to warn the user?
		if (decryptBufferPtr != 0)
		{
			lastBlock = null;
			return new byte[0];
		}

		// lastBlock has the last block we decypted, strip the
		// padding and return it.
		byte[] plain = new byte[BLOCKSIZE - lastBlock[BLOCKSIZE - 1]];

		System.arraycopy(lastBlock, 0, plain, 0, plain.length);

		// tidy up;
		lastBlock = null;

		return plain;
	} // ecbDeFinish()



	/**
	 *	Performs a IDEA encryption.
	 *
	 *	@param plaintext The plain text to encrypt as an array of bytes.
	 *	@return An array of bytes with the cipher text.
 	 *	@see asgaard.utils.cryptography.Cryptography#encrypt
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] encrypt(byte[] plaintext) throws NoSuchMethodException
	{
		// At the moment, only ECB mode is supported.
		return ecbEn(plaintext);
	} // encrypt()



	/**
	 *	Tidies up a IDEA encryption.
	 *
	 *	@return An array of bytes with the cipher text.
 	 *	@see asgaard.utils.cryptography.Cryptography#encryptFinish
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] encryptFinish() throws NoSuchMethodException
	{
		// At the moment, only ECB mode is supported.
		return ecbEnFinish();
	} // encryptFinish()



	/**
	 *	Performs a IDEA decryption.
	 *
	 *	@param ciphertext The cipher text to decrypt, an array of bytes.
	 *	@return An array of bytes with the plain text.
 	 *	@see asgaard.utils.cryptography.Cryptography#decrypt
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] decrypt(byte[] ciphertext) throws NoSuchMethodException
	{
		// At the moment, only ECB mode is supported.
		return ecbDe(ciphertext);
	} // decrypt()



	/**
	 *	Tidies up a IDEA decryption.
	 *
	 *	@return An array of bytes with the plain text.
 	 *	@see asgaard.utils.cryptography.Cryptography#decryptFinish
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] decryptFinish() throws NoSuchMethodException
	{
		// At the moment, only ECB mode is supported.
		return ecbDeFinish();
	} // decryptFinish()



	/**
	 *	Stores the encryption key in the IDEA object.  IDEA keys are
	 *	16 bytes long - if fewer bytes are given, pad with zeros,
	 *	if too many then discard excess.
	 *
	 *	@param key an array of bytes containing the key.
 	 *	@see asgaard.utils.cryptography.Cryptography#setKey
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public void setKey(byte[] key) throws NoSuchMethodException
	{
		// Store away the user key.
		if (key.length >= KEYSIZE)
		{
			System.arraycopy(key, 0, userKey, 0, KEYSIZE);
		}
		else
		{
			System.arraycopy(key, 0, userKey, 0, key.length);
			for (int i = key.length; i < KEYSIZE; i++)
			{
				userKey[i] = (byte)0;
			}
		}

		// Now work out the expanded encryption key.
		expandKey();

		// And from that work out the decryption key.
		invertKey();
	} // setKey()



	/**
	 *	Retrieves the IDEA user key from an object.
	 *
	 *	@return An array of bytes with the user key.
 	 *	@see asgaard.utils.cryptography.Cryptography#getKey
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] getKey() throws NoSuchMethodException
	{
		byte[] key = new byte[KEYSIZE];
		System.arraycopy(userKey, 0, key, 0, KEYSIZE);
		return key;
	} // getKey()
} // class IDEA
