package asgaard.utils.cryptography;

/**
 *	ROT13 is a popular cipher on USEnet News.  Move each alpahbetic
 *	character 13 characters along the alphabet.  Not at all secure,
 *	just used to prevent people accidentally reading offensive material.
 *
 *	Copyright (c) 1997 David Purdue.
 *
 *	@author David Purdue <David.Purdue@Aus.Sun.COM>
 *	@version 1.5 - 97/05/19 14:49:33
 *	@see asgaard.utils.cryptography.Cryptography
 */

public class ROT13 extends Cryptography
{
	/*
	 *	Define constants for a, A, z and Z, since we only rotate
	 *	alphabetic characters.
	 */
	private final static byte CHAR_a = (byte) 'a';
	private final static byte CHAR_z = (byte) 'z';
	private final static byte CHAR_A = (byte) 'A';
	private final static byte CHAR_Z = (byte) 'Z';



	/*
	 *	rotate does the work of ROT13.  We move a character 13 places
	 *	along the alphabet - for this purpose a follows z. Encryption
	 *	and decryption are identical, the lengths of cipher text and
	 *	plain text are the same, and there is no residual to clean up.
	 */
	private byte[] rotate(byte[] plain)
	{
		byte[] cipher = new byte[plain.length];

		for (int i = 0; i < plain.length; i++)
		{
			if ((CHAR_a <= plain[i]) && (plain[i] <= CHAR_z))
			{
				// character is in a - z.
				cipher[i] = (byte)
				((((plain[i] - CHAR_a) + 13) % 26) + CHAR_a);
			}
			else if ((CHAR_A <= plain[i]) && (plain[i] <= CHAR_Z))
			{
				// character is in A - Z.
				cipher[i] = (byte)
				((((plain[i] - CHAR_A) + 13) % 26) + CHAR_A);
			}
			else
			{
				// Not a letter, leave it alone.
				cipher[i] = plain[i];
			}
		}

		return cipher;
	} // rotate()



	/**
	 *	Performs a ROT13 encryption.
	 *
	 *	@param plaintext The plain text to encrypt as an array of bytes.
	 *	@return An array of bytes with the cipher text.
 	 *	@see asgaard.utils.cryptography.Cryptography#encrypt
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] encrypt(byte[] plaintext) throws NoSuchMethodException
	{
		// rotate handles ROT13.
		return rotate(plaintext);
	} // encrypt()



	/**
	 *	Tidies up a ROT13 encryption.
	 *
	 *	@return An array of bytes with the cipher text.
 	 *	@see asgaard.utils.cryptography.Cryptography#encryptFinish
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] encryptFinish() throws NoSuchMethodException
	{
		// No work to do for ROT13.
		byte[] retVal = new byte[0];
		return retVal;
	} // encryptFinish()



	/**
	 *	Performs a ROT13 decryption.
	 *
	 *	@param ciphertext The cipher text to decrypt, an array of bytes.
	 *	@return An array of bytes with the plain text.
 	 *	@see asgaard.utils.cryptography.Cryptography#decrypt
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] decrypt(byte[] ciphertext) throws NoSuchMethodException
	{
		// rotate handles ROT13.
		return rotate(ciphertext);
	} // decrypt()

	/**
	 *	Tidies up a ROT13 decryption.
	 *
	 *	@return An array of bytes with the plain text.
 	 *	@see asgaard.utils.cryptography.Cryptography#decryptFinish
	 *	@exception NoSuchMethodException Thrown by superclass.
	 */
	public byte[] decryptFinish() throws NoSuchMethodException
	{
		// No work to do for ROT13.
		byte[] retVal = new byte[0];
		return retVal;
	} // decryptFinish()
} // class ROT13
