/**
 * This file is part of Skid and Log
 * (P) 1998 by Klaus Hammermller, klaus@ifs.tuwien.ac.at
 * http://www.ifs.tuwien.ac.at/~hammer
 *
 *@version  1.0 170399
 *@autor    Klaus Hammerller
 *@created	170399	KH
 */

package asgaard.utils.log;

// Import the java classes used
import java.io.IOException;
import java.io.File;
import java.lang.Long;
import java.sql.SQLException;
import java.util.Date;
import java.rmi.*;
import java.rmi.server.*;

/**
 * This class implements a simple and robust standalone remote running LOG-server.
 * A wrapper-class for LogBroker with minimum use of code (e.g. no profiles and resources).
 *<p>
 * Before starting <code>asgaard.utils.log.Log</code> the <code><b>rmiregistry</b></code>
 * need to be started!
 *
 *@see asgaard.utils.log.LogBroker
 *@implements asgaard.utils.log.RemoteLog
 */
public class Log extends UnicastRemoteObject implements RemoteLog {

  /**
   * Internal name of the Server (used for Prefixes and generated messages)
   */
  public static final String SERVER = "LOG";
  /**
   * Version of the Implementation
   */
  public static final String VERSION = "1.0.2";
  /**
   * The Copyright message
   */
  public static final String COPYRIGHT = "(P) by the Asgaard Project group http://www.ifs.tuwien.ac.at/asgaard\n"+
                                         "    under GNU public license - absolute NO WARRANTY on this software.";
  /**
   * Binding-Name of the LOG-Server within the <code>rmiregistry</code>
   */
  public static String ServerName = "LogServer";        // Binding Name
  private static LogBroker log = null;                  //The Log-Broker
  private static boolean debug = false;

  private static byte MailLevel = 1;                    // Which Statements shall be mailed too
  private static String LOGpath = ".log";               // Path to the LOG-file
  private static long LOGsize = 0;                      // Size of the Logfile until a warning occours
  private static String LOGmailTO = null;               // Mail to
  private static String LOGmailCC = "";
  private static String LOGmailHOST = null;

  /**
   * Instanciates a new server-object and sets the local path. If available also the mailserver is instanciated.
   */
  public Log( String path ) throws RemoteException {
    try {
      log = new LogBroker( path );
      if ((LOGmailTO != null) && (LOGmailHOST != null))
        log.setMailServer(MailLevel, LOGmailTO, ServerName, LOGmailCC, LOGmailHOST, LOGsize);
    } catch (Exception e) {
      System.err.println(ServerName + ": " + e.toString() );
      if (debug) e.printStackTrace();                     
      System.exit( 1 );
    }
  }

  /**
   * Puts a Statement into the LOGFILE (wrapper method to LogBroker).
   * If the Statement exists already, it is replaced.
   * If any Exception rises it will be ignored
   *
   *@see asgaard.utils.log.LogMsg
   *@see asgaard.utils.log.LogBroker#put
   *@param stmt Statement as a LogMsg-Bean
   *@exception SQLException from LogBroker.put forwarded exception
   *@exception IOException from LogBroker.put forwarded exception
   *@exception RemoteException if the remote (only if remote!) Object cannot be accessed
   */
  public void put(LogMsg stmt) throws RemoteException, IOException, SQLException {
    log.put( stmt );
  }

  /**
   * Gets a Statement from the LOGFILE returning the Statement-bean (wrapper method to LogBroker).
   *
   *@see asgaard.utils.log.LogMsg
   *@see asgaard.utils.log.LogBroker#get
   *@param nr Number of the statement in the current Logfile
   *@param withBean if the Bean carried by the Statement is required
   *@return a List with the found Statement-Beans
   *@exception RemoteException if the remote (only if remote!) Object cannot be accessed
   */
  public LogMsg[] get(long nr, boolean withBean) throws RemoteException {
    return log.get( nr, withBean );
  }

  /**
   * Defines a Mailserver for automatically forward LOGMSG also via eMail.
   *
   *@param level Which Error-Level should be forwarded
   *@param to mailto - eMailadress
	 *@param from eMailadress
	 *@param cc carbon copy (optional)
	 *@param mailhost assumption: mail.smtp.host
   *@param logsize send warning if logfile exceeds size (0 for no warning)
   *@exception RemoteException if the remote (only if remote!) Object cannot be accessed
   */
  public void setMailServer(byte level, String to, String from, String cc, String mailhost, long size) throws RemoteException {
    log.setMailServer(level, to, from, cc, mailhost, size);
  }
  /**
   * Starting the remote Server.
   *@param args commandline parameter
   */
  public static void main(String[] args) {
    System.setSecurityManager( new RMISecurityManager() );    // Create and install the security Server
    try {
        System.out.println("Hello, this is " + SERVER + " V" + VERSION);
        System.out.println(COPYRIGHT);
                                                          // --------------------------------
		    for (int i = 0; i < args.length; i++) {           // ANALYZING THE COMMANDLINE_PARAMs
			    if (args[i].charAt(0) == '-') {
				    boolean skiparg = false;
    				int j = 1;
	    			while ((j < args[i].length()) && (!skiparg)) {
		    			switch(args[i].charAt(j)) {
			    		case 'h': case '?':                         // HELP
                usage(true); break;
              case 'm':                                   // MailLevel
                switch(args[i].charAt(j+1)) {
                  case '0': MailLevel = -1; j +=1; break;
                  case '1': MailLevel = 0; j +=1; break;
                  case '2': MailLevel = 2; j +=1; break;
                  case '3': MailLevel = 2; j +=1; break;
                  case '4': MailLevel = 3; j +=1; break;
                  case '5': MailLevel = 4; j +=1; break;
                  case '6': MailLevel = 5; j +=1; break;
					    		default: usage(false); break;
                } break;
  				  	case 'p':                                   // PATH
	  				  	if ((i+1 >= args.length) || (j+1 == args[i].length()))
		  				  	usage(false);
  			  			else {
	  			  			String tstr = null;
		  			  		if (j+2 >= args[i].length()) {
			  			  		tstr = args[i+1];
				  			  	skiparg = true;
  				  			} else
	  				  			tstr = args[i].substring(j+2, args[i].length());
  			  				switch (args[i].charAt(j+1)) {
	      						case 'H': LOGmailHOST = tstr; break;
                    case 'L': if (!tstr.endsWith(File.separator)) tstr += File.separator;
                              LOGpath = tstr; break;
	      						case 'M': LOGmailCC = tstr; break;
	      						case 'T': LOGmailTO = tstr; break;
	      						case 'X': LOGsize = Long.getLong(tstr).longValue() * 1024; break;
  					    		default: usage(false); break;
	  						  }
    	  					if (!skiparg) j = args[i].length()-1;
		  	  			}
			  	  		break;
				  	  default: usage(false); break;
  					  }
  	  				j++;
	  	  		}
		  	  	if (skiparg) i++;
          }
  			}
      Log app = new Log( LOGpath );                           // Instanciate the LOG-server
      Naming.rebind( ServerName, app );                       // binding the Name in the registry
      app.invokedStandalone = true;
      Date now = new Date ();                                 // Creating a Statement
      log.put( new LogMsg( SERVER, "main#1",
                          LogMsg.LOG_REMARK, true,
                          ServerName + " V" + VERSION + " started successfuly."));
      System.out.println("Logserver started at " + now.toString() +
                         " bound in the registry to the name " + ServerName);
    } catch (Exception e) {
      System.err.println(ServerName + ": " + e.toString());
      if (debug) e.printStackTrace();
      try {
        if (log != null)
          log.put(new LogMsg( SERVER, "main#2",
                              LogMsg.LOG_ERROR, false,
                              ServerName + " V" + VERSION + " failed to start."));
      } catch (Exception f) { }
      System.exit( 1 );
    }
  }
  private boolean invokedStandalone = false;

  // usage of the Logserver

  private static void usage( boolean longusage) {
		System.out.println( "Usage: java asgaard.utils.log.Log [-hpLHMTm]" );
    if (longusage)
		  System.out.println( "\t-h\t\t\tprint this help message\n"+
		    "\t-pL <LOGpath>\t\tlocal directory to write the LOGFILE\n"+
		    "\t-pH <mail HOST>\t\tused mailer (local log only)to\n"+
		    "\t-pM <mail CC>\t\tcc mailed statements (local log only)\n"+
		    "\t-pT <mail TO>\t\tmailed statements to (local log only)\n"+
		    "\t-pK <LOGsize>\t\tmail if the Logfile extends <LOGsize>kb\n"+
		    "\t-m<statement>\t\ttype of statements which are forwarded\n"+
		    "\t\t\t0\tforward no statements\n"+
		    "\t\t\t1\tforward error statements\n"+
		    "\t\t\t2\tforward warning statements\n"+
		    "\t\t\t3\tforward remark statements\n"+
		    "\t\t\t4\tforward backup statements\n"+
		    "\t\t\t5\tforward debug statements");
    System.exit( 0 );
  }
}