/**
 * This file is part of Skid and Log
 * (P) 1998 by Klaus Hammermller, klaus@ifs.tuwien.ac.at
 * http://www.ifs.tuwien.ac.at/~hammer
 *
 *@version  1.0 170399
 *@autor    Klaus Hammerller
 *@created	170399	KH
 */

package asgaard.utils.log;

// Import the java classes used
import java.util.*;
import java.io.IOException;
import java.sql.SQLException;
import java.net.*;
import java.rmi.*;

/**
 * LogClient is a wrapper for local and remote Logserver and does the Exception-handling.
 * To guarantee a minimum performance this is implemented as <code>Thread</code> which can be used or not as a <code>Thread</code>.
 */
public class LogClient extends Thread {

  private RemoteLog log;                  // LogBroker
  private String prefix;
  private Vector putStmt = new Vector();  // Hold the Stmts to process
  private boolean done = false;
  private boolean debug = false;
  private boolean remoteLog = false;
  private boolean thread = false;
  private String hostName = LogMsg.LOCAL_HOST;

  /**
   * Creates a local LogClient, wrapping a local Logfile with the default directory.
   *
   *@param pre  prefix printed to System.err in case of an Exception
   *@param start runs the LogClient as Thread
   *@exception SQLException if the JDBC connection to the <i>local</i> flatfile (logfile) failes
   *@exception IOException if the access to the <i>local</i> flatfile (logfile) failes
   */
  public LogClient(String pre, boolean start) throws IOException, SQLException {
    prefix = pre;
    try {
      InetAddress local = InetAddress.getLocalHost(); // name of the Host the requesting Server is located
      hostName = local.getHostName() + " ["+local.getHostAddress()+"]";
    } catch (java.net.UnknownHostException e) {  }
    log = new LogBroker( );
    if (start) {
      thread = true;
      this.setPriority(Math.max(this.getPriority()-1, Thread.MIN_PRIORITY));
      start();                                              // start the LogClient
    }
  }
  /**
   * Creates a local LogClient, wrapping a local Logfile and setup a mailclient.
   *
   *@param pre  prefix printed to System.err in case of an Exception
   *@param localPath  path for the local logfile
   *@param mailLevel  mail statements up to specified level
   *@param mailTO adress of to send the statements
   *@param clientName name of the client-task (\@[host] is added automatically)
   *@param mailCC adress of to send the statements
   *@param mailHOST mail host
   *@param size maximum size of the logfile
   *@param start runs the LogClient as Thread
   *@exception SQLException if the JDBC connection to the <i>local</i> flatfile (logfile) failes
   *@exception IOException if the access to the <i>local</i> flatfile (logfile) failes
   */
  public LogClient(String pre, String localPath, byte mailLevel, String mailTO, String clientName, String mailCC, String mailHOST, long size, boolean start) throws IOException, SQLException{
    prefix = pre;
    try {
      InetAddress local = InetAddress.getLocalHost(); // name of the Host the requesting Server is located
      hostName = local.getHostName() + " ["+local.getHostAddress()+"]";
    } catch (java.net.UnknownHostException e) {  }
    log = new LogBroker(localPath);
    if ((mailTO != null) && (mailHOST != null)) {
      InetAddress local = InetAddress.getLocalHost(); // name of the Host the requesting Server is located
      log.setMailServer(mailLevel, mailTO, clientName + "@" + local.getHostName(), mailCC, mailHOST, size);
    }
    if (start) {
      thread = true;
      this.setPriority(Math.max(this.getPriority()-1, Thread.MIN_PRIORITY));
      start();                                              // start the LogClient
    }
  }
  /**
   * Creates a local LogClient, wrapping a <i>remote</i> Logserver.
   * If the connection to the remote Logserver failes a default local Log is created.
   *
   *@param pre  prefix printed to System.err in case of an Exception
   *@param server name of the Logserver host
   *@param start runs the LogClient as Thread
   *@exception SQLException if the JDBC connection to the <i>local</i> flatfile (logfile) failes (backup routine after remote failue)
   *@exception IOException if the access to the <i>local</i> flatfile (logfile) failes (backup routine after remote failue)
   */
  public LogClient(String pre, String server, boolean start) throws IOException, SQLException {
    prefix = pre;
    try {                                                 // llok for remote Logserver
      log = (RemoteLog) Naming.lookup("//" + server + "/" + Log.ServerName );
      remoteLog = true;
    } catch (Exception f) {
      System.err.println(pre + f.toString()+ " " + f.getMessage());
      if (debug) f.printStackTrace();
      log = new LogBroker( );                             // Try a local logfile
      put( this.toString(), "LogClient#1",                // Create a statement
           LogMsg.LOG_WARNING, true,
           " Could not connect to //" + server + "/" + Log.ServerName + ", created local logfile " + LogBroker.local_path + LogBroker.LOG_FILE_NAME + LogBroker.LOG_FILE_EXTENSION);
    }
    if (start) {
      thread = true;
      this.setPriority(Math.max(this.getPriority()-1, Thread.MIN_PRIORITY));
      start();                                              // start the LogClient
    }
  }

  /**
   * Does the work for put as a Thread
   */
  public void run() {
    Object obj = null;
    LogMsg stmt = null;
    Object[] param;

    while (!done) {
      if (!thread) done = true;
      try {
        if (!putStmt.isEmpty()) {           // Get the first Element from the Vector
          obj = putStmt.firstElement();
          putStmt.removeElementAt( 0 );
          if (obj instanceof LogMsg)
            stmt = (LogMsg) obj;
          else
            if (obj instanceof Object[]) {
              param = (Object[]) obj;
              if (param.length == 6)
                stmt = new LogMsg(  hostName,
                                    (String)param[0],
                                    (String)param[1],
                                    ((Byte)param[2]).byteValue(),
                                    ((Boolean)param[3]).booleanValue(),
                                    (String)param[4],
                                    param[5]);
            }
          if (!(stmt==null))                // put the statement
            log.put( stmt );
          stmt = null;
        } else
          this.suspend();
      } catch (Exception e) {
        stmt = null;
        System.err.println(prefix + e.toString() + " " + e.getMessage());
        if (debug) e.printStackTrace();
      }
    }
    if (!thread) done = false;
  }

  /**
   * Defines a Mailserver for automatically forward LOGMSG also via eMail.
   *
   *@param level Which Error-Level should be forwarded
   *@param to mailto - eMailadress
	 *@param from eMailadress
	 *@param cc carbon copy (optional)
	 *@param mailhost assumption: mail.smtp.host
   *@param logsize send warning if logfile exceeds size (0 for no warning)
   */
  public void setMailServer(byte level, String to, String from, String cc, String mailhost, long size) {
    try {
      log.setMailServer(level, to, from, cc, mailhost, size);
    } catch (Exception e) {
      System.err.println(prefix + e.toString() + " " + e.getMessage());
      if (debug) e.printStackTrace();
    }
  }

  /**
   * Puts a Statement into the wrapped LOGFILE.
   * If the Statement exists already, it is replaced.
   * If any Exception rises it will be ignored
   *
   *@see asgaard.utils.log.LogMsg
   *@param stmt Statement as a LogMsg-Bean
   */
  public void put(LogMsg stmt) {
    putStmt.addElement( stmt );
    if (!thread)
      run();
    else
      this.resume();
  }

  /**
   * Creates a Statement out of the Message
   * If any Exception rises it will be ignored
   *
   *@see asgaard.utils.log.LogMsg
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   */
  public void put(String server, String point, byte typ, boolean ok, String msg) {
    Object[] param = new Object[6];
    param[0] = server;
    param[1] = point;
    param[2] = new Byte( typ );
    param[3] = new Boolean( ok );
    param[4] = msg;
    param[5] = null;
    putStmt.addElement( param );
    if (!thread)
      run();
    else
      this.resume();
  }

  /**
   * Creates a Statement out of the Message
   * If any Exception rises it will be ignored
   *
   *@see asgaard.utils.log.LogMsg
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   *@param bean       Some streamable Object
   */
  public void put(String server, String point, byte typ, boolean ok, String msg, Object bean) {
    Object[] param = new Object[6];
    param[0] = server;
    param[1] = point;
    param[2] = new Byte( typ );
    param[3] = new Boolean( ok );
    param[4] = msg;
    param[5] = bean;
    putStmt.addElement( param );
    if (!thread)
      run();
    else
      this.resume();
  }
  /**
   * Creates a Backup-Statement out of the Message
   * If any Exception rises it will be ignored
   *
   *@see asgaard.utils.log.LogMsg
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   *@param bean       Some streamable Object
   */
  public void putBackup(String server, String point, String msg, Object bean) {
    Object[] param = new Object[6];
    param[0] = server;
    param[1] = point;
    param[2] = new Byte( LogMsg.LOG_BACKUP );
    param[3] = new Boolean( true );
    param[4] = msg;
    param[5] = bean;
    putStmt.addElement( param );
    if (!thread)
      run();
    else
      this.resume();
  }
  /**
   * Gets a Statement from the wrapped LOGFILE returning the Statement-bean.
   *
   *@see asgaard.utils.log.LogMsg
   *@param nr Number of the statement in the current Logfile
   *@param withBean if the Bean carried by the Statement is required
   *@return a List with the found Statement-Beans
   */
  public LogMsg[] get(long nr, boolean withBean) {
    try {
      return log.get( nr, withBean );
    } catch (Exception e) {
      System.err.println(prefix + e.toString() + " " + e.getMessage());
      if (debug) e.printStackTrace();
      return null;
    }
  }

  /**
   * Is the current Logfile remote?
   */
  public boolean isRemote() {
    return remoteLog;
  }

}