/**
 * This file is part of Skid and Log
 * (P) 1998 by Klaus Hammermller, klaus@ifs.tuwien.ac.at
 * http://www.ifs.tuwien.ac.at/~hammer
 *
 *@version  1.0 270299
 *@autor    Klaus Hammerller
 *@created	270299	KH
 */

package asgaard.utils.log;

// Import the java classes used
import java.io.*;
import java.util.*;
import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * This is a Bean that carries a statement in the LOG - File.
 * Collection and Distribution of statements within the <code>LogBroker</code>.
 *
 *@see asgaard.utils.log.LogBroker
 */
public class LogMsg implements Serializable {

  /**
   * Log-level "No Message"
   */
  public  final static byte LOG_NONE = -1;         // VALUES DEFINING THE LogType

  /**
   * Log-level "Fatal Error", an error occoured and can not be recovered automaticaly
   */
  public  final static byte LOG_SECURITY = 0;
  /**
   * Log-level "Security violation", occour if an anourmalous behavior is detected.
   */
  public  final static byte LOG_ERROR = 1;
  /**
   * Log-level "Profile", storing profiling- and evaluation information 
   */
  public  final static byte LOG_PROFILE = 2;
  /**
   * Log-level "Warning", an error occoured and has been repaired automatically but may not work in the intended way (e.g. locally but remote)
   */
  public  final static byte LOG_WARNING = 3;
  /**
   * Log-level "Remark" by the Server as documentation of the Server-State
   */
  public  final static byte LOG_REMARK = 4;
  /**
   * Log-level "Backup" Statements to ensave information
   */
  public  final static byte LOG_BACKUP = 5;
  /**
   * Log-level "Debuginformation" to trace processes about different server-tasks
   */
  public  final static byte LOG_DEBUG = 6;

  private final static byte LOG_MIN = LOG_SECURITY;
  private final static byte LOG_MAX = LOG_DEBUG;

  /**
   * No Bean included to the Statement
   */
  public  final static byte BEAN_NONE = 0;         // VALUES DEFINING THE additional BEAN
  /**
   * Bean available at the LogBroker but not included
   */
  public  final static byte BEAN_NOT_INCLUDED = 1;
  /**
   * Some complex bean which is some Object
   */
  public  final static byte BEAN_OBJECT = 2;
  /**
   * Bean is a simple type and stored as String
   */
  public  final static byte BEAN_STRING = 3;

  private final static byte BEAN_MIN = BEAN_NONE;
  private final static byte BEAN_MAX = BEAN_STRING;

  /**
   * If the Hostname cannot be resolved
   */
  public final static String LOCAL_HOST = "LOCAL";

  private long    LogNr = 0;                      // Number of the Message within the logfile
  private String  HostName = null;                // name of the Host the requesting Server is located
  private String  ServerClass = null;             // The calling Java-Serverclass
  private String  ServicePoint = null;            // The Method / place where the statement was created
  private byte    LogType = 0;                    // The Type of the LOG-Statement
  private Date    Created = new Date(System.currentTimeMillis() );        // Local Time when the Statement was created
  private Date    Received = Created;             // Local Time when the Statement was processed by the LOG-server
  private boolean success = false;                // Was the Action successful (only for Remark, Backup and Debug - Types)
  private String  LogMessage = null;              // Human redable Message of the Statement
  private byte    BeanType = 0;                   // The Type of the stored Statement
  private Object  LogBean = null;                 // Some streamable Object

  /**
   * Default-constructor, do not use manually - <b>internal use only!</b> Only used for Streaming.
   */
  public LogMsg() { }

  /**
   * Manual constructor for a short statement without included streamable object (Bean)
   *
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   *@exception IllegalArgumentException if the parameter <code>typ</code> does not fit to the log-types defined
   */
  public LogMsg(String server, String point, byte typ, boolean ok, String msg)  throws IllegalArgumentException, NotSerializableException {
    this(server, point, typ, ok, msg, null);
  }
  /**
   * Manual constructor for a short statement including a streamable object (Bean)
   *
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   *@param bean       Some streamable Object
   *@exception IllegalArgumentException if the parameter <code>typ</code> does not fit to the log-types defined
   *@exception NotSerializableException if the parameter <code>bean</code> can not be serialized
   */
  public LogMsg(String server, String point, byte typ, boolean ok, String msg, Object bean)  throws IllegalArgumentException, NotSerializableException {
    try {
      InetAddress local = InetAddress.getLocalHost(); // name of the Host the requesting Server is located
      setHostName(local.getHostName() + " ["+local.getHostAddress()+"]");
    } catch (UnknownHostException e) {
      setHostName( LOCAL_HOST );
    }
    setServerClass(server);                         // The calling Java-Serverclass
    setServicePoint(point);                         // The Method / place where the statement was created
    setLogType(typ);                                // The Type of the LOG-Statement
    setSuccess( ok );
    setLogMessage(msg);                             // Human redable Message of the Statement
    if (!(bean==null))
      setLogBean(bean);                             // Some streamable Object
  }

  /**
   * Manual and automated constructor for a logclients (used in LogClient).
   *
   *@param host       The host where the statement was created
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   *@param bean       Some streamable Object
   *@exception IllegalArgumentException if the parameter <code>typ</code> does not fit to the log-types defined
   *@exception NotSerializableException if the parameter <code>bean</code> can not be serialized
   */
  public LogMsg(String host, String server, String point, byte typ, boolean ok, String msg, Object bean)  throws IllegalArgumentException, NotSerializableException {
    setHostName( host );
    setServerClass(server);                         // The calling Java-Serverclass
    setServicePoint(point);                         // The Method / place where the statement was created
    setLogType(typ);                                // The Type of the LOG-Statement
    setSuccess( ok );
    setLogMessage(msg);                             // Human redable Message of the Statement
    if (!(bean==null))
      setLogBean(bean);                             // Some streamable Object
  }

  /**
   * Complete constructor for a historical Statement, do not use manually - </b>internal use only!</b> (LogBroker)
   *
   *@see asgaard.utils.log.LogBroker
   *@param nr         Number of the statement
   *@param host       name of the Host the requesting Server is located
   *@param server     The calling Java-Serverclass
   *@param point      The Method / place where the statement was created
   *@param c          Timestamp when the statement was created
   *@param d          Timestamp when the statement was received
   *@param typ        The Type of the LOG-Statement
   *@param ok         Was the ServicePoint passed successfuly
   *@param msg        Human redable Message of the Statement
   *@param bean       Some streamable Object
   *@param beanExist  only used, if bean=null and beantype=BEAN_NOT_INCLUDED)
   *@exception IllegalArgumentException if the parameter <code>typ</code> does not fit to the log-types defined
   *@exception NotSerializableException if the parameter <code>bean</code> can not be serialized
   */
  public LogMsg(long nr, String host, String server, String point, Date c, Date d, byte typ, boolean ok, String msg, Object bean, boolean beanExist)  throws IllegalArgumentException, NotSerializableException {
    setHostName(host);                    // name of the Host the requesting Server is located
    setServerClass(server);               // The calling Java-Serverclass
    setServicePoint(point);               // The Method / place where the statement was created
    setLogType(typ);                      // The Type of the LOG-Statement
    setSuccess( ok );
    setLogMessage(msg);                   // Human redable Message of the Statement
    if (bean == null) {
      if (beanExist) BeanType = BEAN_NOT_INCLUDED;
    } else
      setLogBean(bean);                         // Some streamable Object
    Created = new java.util.Date(c.getTime());  // Set the timestamps too
    setReceived( d );
    setLogNr( nr );                       // Set the Number of the Message
  }

  /**
   * Set the number of the Message within the logfile, do not use manually - <b>internal use only!</b> (LogBroker)
   */
  public void setLogNr(long nr) {
    LogNr = nr;
  }
  /**
   * Set the name of the Host the requesting Server is located, default is the localhost.
   */
  public void setHostName(String name) {
    HostName = name;
  }
  /**
   * Set the calling Java-Serverclass
   */
  public void setServerClass (String name) {
    ServerClass = name;
  }
  /**
   * Set the Method / place where the statement was created
   */
  public void setServicePoint (String name) {
    ServicePoint = name;
  }
  /**
   * Set the Type of the LOG-statement ("Log-level")
   *@exception IllegalArgumentException if the parameter does not fit to the log-types defined
   */
  public void setLogType(byte val) throws IllegalArgumentException {
    if ((val < LOG_MIN) | (val > LOG_MAX))        // Rangecheck
      throw new IllegalArgumentException("Wrong LogType");
    LogType = val;
  }

  /**
   * Set the Local Time when the Statement was processed by the LOG-server
   */
  public void setReceived (java.util.Date ts) {
    Received = new java.util.Date(ts.getTime());
  }
  /**
   * Set if the Servicepoint was passed successfuly.
   */
  public void setSuccess(boolean ok) {
    success = ok;
  }
  /**
   * Set the message to the statement
   */
  public void setLogMessage (String msg) {
    LogMessage = msg;
  }
  /**
   * Set some streamable object
   *@param bean the Bean which shall be included to the message, <code>null</code> setzt the <code>BeanType</code> to <code>BEAN_NONE</code>.
   *
   *@exception NotSerializableException if the parameter can not be serialized (<code>null</code> is ignored)
   */
  public void setLogBean  (Object bean) throws NotSerializableException {
    if (bean == null) {                                     // Check if Bean exists
      BeanType = BEAN_NONE;
      return;
    }
    if (bean instanceof java.lang.String) {                 // Take it as a String
      BeanType = BEAN_STRING;
      LogBean = bean;
    } else {
      if (!(bean instanceof Serializable)) {
	      String clname = bean.getClass().getName();
        throw new NotSerializableException(clname);
      }
      LogBean = bean;                                       // set property
      BeanType = BEAN_OBJECT;
    }
  }

  /**
   * Get number of the Message within the logfile
   */
  public  long getLogNr() { return LogNr; }
  /**
   * Get name of the Host the requesting Server is located
   */
  public String getHostName() { return HostName; }
  /**
   * Get the calling Java-Serverclass
   */
  public String getServerClass() { return ServerClass; }
  /**
   * Get the Method / place where the statement was created
   */
  public String getServicePoint() { return ServicePoint; }
  /**
   * Get the Type of the LOG-Statement
   */
  public byte getLogType() { return LogType; }
  /**
   * Converts the Type of the LOG-Statement
   *
   *@param the asked type
   *@return a String
   */
  public static String getLogType(byte type) {
    switch (type) {
      case LOG_SECURITY: return "SECURITY";
      case LOG_ERROR: return "ERROR";
      case LOG_PROFILE: return "PROFILE";
      case LOG_WARNING: return "WARNING";
      case LOG_REMARK: return "REMARK";
      case LOG_BACKUP: return "BACKUP";
      case LOG_DEBUG: return "DEBUG";
      default: return "NONE";
    }
  }
  /**
   * Converts the Type of the LOG-Statement
   *
   *@param the asked type
   *@return a String
   */
  public static byte getLogType(String type) {
    if (type == null) return LOG_NONE;
    char first = type.charAt( 0 );
    switch (first) {
      case 'E': case 'e': return LOG_ERROR;
      case 'P': case 'p': return LOG_PROFILE;
      case 'W': case 'w': return LOG_WARNING;
      case 'R': case 'r': return LOG_REMARK;
      case 'B': case 'b': return LOG_BACKUP;
      case 'D': case 'd': return LOG_DEBUG;
      default: return LOG_NONE;
    }
  }
  /**
   * Get Local Time when the Statement was created, precision "seconds/1000"
   */
  public Date getCreated() { return Created; }
  /**
   * Get Local Time when the Statement was processed by the LOG-server, precision "seconds"
   */
  public Date getReceived() { return Received; }
  /**
   * Get was the Action successful
   */
  public boolean isSuccess() { return success; }
  /**
   * Get was the Action successful
   */
  public String getOk() { return getOk (success); }
  /**
   * Converts <code>boolean</code> to String
   */
  public static String getOk(boolean ok) {
    if (ok) return "ok"; else return "fail";
  }
  /**
   * Converts String to <code>boolean</code>
   */
  public static boolean getOk(String ok) {
    if (ok == null) return false;
    if ((ok.charAt(0) == 'o') || (ok.charAt(0) == 'O')) return true;
    return false;
  }
  /**
   * Get human redable Message of the Statement
   */
  public String getLogMessage() { return LogMessage; }
  /**
   * Get some streamable Object
   */
  public Object getLogBean() { return LogBean; }
  /**
   * Get the type of the Bean included to the Statement
   */
  public byte getBeanType() { return BeanType; }
}
